class Firmware:
    def __init__(self, firmware_path: str, ipsw):
        self._ipsw = ipsw
        self._firmware_path = firmware_path
        self._manifest_data = self._ipsw.read(self.get_relative_path('manifest'))
        self._firmware_files = {}
        for filename in self._manifest_data.splitlines():
            filename = filename.strip()
            component_name = self.get_component_name(filename)
            self._firmware_files[component_name] = f'{firmware_path}/{filename}'

    def get_relative_path(self, path: str):
        return f'{self._firmware_path}/{path}'

    def get_files(self):
        return self._firmware_files

    @staticmethod
    def get_component_name(filename):
        names = {
            'LLB': 'LLB',
            'iBoot': 'iBoot',
            'DeviceTree': 'DeviceTree',
            'applelogo': 'AppleLogo',
            'liquiddetect': 'Liquid',
            'lowpowermode': 'LowPowerWallet0',
            'recoverymode': 'RecoveryMode',
            'batterylow0': 'BatteryLow0',
            'batterylow1': 'BatteryLow1',
            'glyphcharging': 'BatteryCharging',
            'glyphplugin': 'BatteryPlugin',
            'batterycharging0': 'BatteryCharging0',
            'batterycharging1': 'BatteryCharging1',
            'batteryfull': 'BatteryFull',
            'needservice': 'NeedService',
            'SCAB': 'SCAB',
            'sep-firmware': 'RestoreSEP',
        }
        return names.get(filename)



import logging
import zipfile
from contextlib import contextmanager

from cached_property import cached_property
from construct import Const, Default, PaddedString, Struct

from pymobiledevice3.restore.ipsw.build_manifest import BuildManifest
from pymobiledevice3.restore.ipsw.firmware import Firmware

cpio_odc_header = Struct(
    'c_magic' / Const('070707', PaddedString(6, 'utf8')),
    'c_dev' / Default(PaddedString(6, 'utf8'), '0' * 6, ),
    'c_ino' / PaddedString(6, 'utf8'),
    'c_mode' / PaddedString(6, 'utf8'),
    'c_uid' / Default(PaddedString(6, 'utf8'), '0' * 6),
    'c_gid' / Default(PaddedString(6, 'utf8'), '0' * 6),
    'c_nlink' / PaddedString(6, 'utf8'),
    'c_rdev' / Default(PaddedString(6, 'utf8'), '0' * 6),
    'c_mtime' / Default(PaddedString(11, 'utf8'), '0' * 11),
    'c_namesize' / PaddedString(6, 'utf8'),
    'c_filesize' / Default(PaddedString(11, 'utf8'), '0' * 11),
)


class IPSW:
    def __init__(self, archive: zipfile.ZipFile):
        self._archive = archive
        self.logger = logging.getLogger(__file__)
        self.build_manifest = BuildManifest(self, self._archive.read('BuildManifest.plist'))

    @cached_property
    def restore_version(self):
        return self.read('RestoreVersion.plist')

    @cached_property
    def system_version(self):
        return self.read('SystemVersion.plist')

    @cached_property
    def filelist(self):
        return self._archive.filelist

    @contextmanager
    def open_path(self, path: str):
        file = self._archive.open(path)
        try:
            yield file
        finally:
            file.close()

    @property
    def bootability(self):
        result = b''
        prefix = 'BootabilityBundle/Restore/Bootability/'
        inode = 1
        nlink = 1

        for e in self.filelist:
            if e.filename == 'BootabilityBundle/Restore/Firmware/Bootability.dmg.trustcache':
                subpath = 'Bootability.trustcache'
            elif not e.filename.startswith(prefix):
                continue
            else:
                subpath = e.filename[len(prefix):]

            self.logger.debug(f'BootabilityBundle: adding {subpath}')

            filename = subpath
            filename = f'{filename}\0'.encode()
            mode = e.external_attr >> 16
            result += cpio_odc_header.build({
                'c_ino': f'{inode:06o}', 'c_nlink': f'{nlink:06o}', 'c_mode': f'{mode:06o}',
                'c_namesize': f'{len(filename):06o}', 'c_filesize': f'{e.file_size:011o}'})
            inode += 1
            result += filename
            if not e.file_size:
                continue

            with self.open_path(e.filename) as f:
                result += f.read()

        filename = b'TRAILER!!!\0'
        inode = 0
        mode = 0
        result += cpio_odc_header.build(
            {'c_ino': f'{inode:06o}', 'c_mode': f'{mode:06o}', 'c_nlink': f'{nlink:06o}',
             'c_namesize': f'{len(filename):06o}'}) + filename
        return result

    def read(self, path: str):
        return self._archive.read(path)

    def get_global_manifest(self, macos_variant: str, device_class: str):
        manifest_path = f'Firmware/Manifests/restore/{macos_variant}/apticket.{device_class}.im4m'
        return self.read(manifest_path)

    def get_firmware(self, firmware_path: str):
        return Firmware(firmware_path, self)



import abc
import plistlib
import socket
import sys
import time
from construct import Struct, Prefixed, Int32ul, GreedyBytes, StreamError, Int16ul, CString, Padding, FixedSized, \
    Enum, Const, Switch, this
from dataclasses import dataclass
from pymobiledevice3.exceptions import MuxException, MuxVersionError, NotPairedError, UsbmuxConnectionError
from typing import List, Optional, Mapping

usbmuxd_version = Enum(Int32ul,
                       BINARY=0,
                       PLIST=1,
                       )

usbmuxd_result = Enum(Int32ul,
                      OK=0,
                      BADCOMMAND=1,
                      BADDEV=2,
                      CONNREFUSED=3,
                      BADVERSION=6,
                      )

usbmuxd_msgtype = Enum(Int32ul,
                       RESULT=1,
                       CONNECT=2,
                       LISTEN=3,
                       ADD=4,
                       REMOVE=5,
                       PAIRED=6,
                       PLIST=8,
                       )

usbmuxd_header = Struct(
    'version' / usbmuxd_version,  # protocol version
    'message' / usbmuxd_msgtype,  # message type
    'tag' / Int32ul,  # responses to this query will echo back this tag
)

usbmuxd_request = Prefixed(Int32ul, Struct(
    'header' / usbmuxd_header,
    'data' / Switch(this.header.message, {
        usbmuxd_msgtype.CONNECT: Struct(
            'device_id' / Int32ul,
            'port' / Int16ul,  # TCP port number
            'reserved' / Const(0, Int16ul),
        ),
        usbmuxd_msgtype.PLIST: GreedyBytes,
    }),
), includelength=True)

usbmuxd_device_record = Struct(
    'device_id' / Int32ul,
    'product_id' / Int16ul,
    'serial_number' / FixedSized(256, CString('ascii')),
    Padding(2),
    'location' / Int32ul
)

usbmuxd_response = Prefixed(Int32ul, Struct(
    'header' / usbmuxd_header,
    'data' / Switch(this.header.message, {
        usbmuxd_msgtype.RESULT: Struct(
            'result' / usbmuxd_result,
        ),
        usbmuxd_msgtype.ADD: usbmuxd_device_record,
        usbmuxd_msgtype.REMOVE: Struct(
            'device_id' / Int32ul,
        ),
        usbmuxd_msgtype.PLIST: GreedyBytes,
    }),
), includelength=True)


@dataclass
class MuxDevice:
    devid: int
    serial: str
    connection_type: str

    def connect(self, port) -> socket.socket:
        return create_mux().connect(self, port)

    @property
    def is_usb(self) -> bool:
        return self.connection_type == 'USB'

    @property
    def is_network(self) -> bool:
        return self.connection_type == 'Network'

    def matches_udid(self, udid: str) -> bool:
        return self.serial.replace('-', '') == udid.replace('-', '')


class SafeStreamSocket:
    """ wrapper to native python socket object to be used with construct as a stream """

    def __init__(self, address, family):
        self.sock = socket.socket(family, socket.SOCK_STREAM)
        self.sock.connect(address)

    def send(self, msg: bytes) -> int:
        self.sock.sendall(msg)
        return len(msg)

    def recv(self, size: int) -> bytes:
        msg = b''
        while len(msg) < size:
            chunk = self.sock.recv(size - len(msg))
            if not chunk:
                raise MuxException('socket connection broken')
            msg += chunk
        return msg

    def close(self):
        self.sock.close()

    def settimeout(self, interval: float):
        self.sock.settimeout(interval)

    def setblocking(self, blocking: bool):
        self.sock.setblocking(blocking)

    read = recv
    write = send


class MuxConnection:
    # used on Windows
    ITUNES_HOST = ('127.0.0.1', 27015)

    # used for macOS and Linux
    USBMUXD_PIPE = '/var/run/usbmuxd'

    @staticmethod
    def create_usbmux_socket() -> SafeStreamSocket:
        try:
            if sys.platform in ['win32', 'cygwin']:
                return SafeStreamSocket(MuxConnection.ITUNES_HOST, socket.AF_INET)
            else:
                return SafeStreamSocket(MuxConnection.USBMUXD_PIPE, socket.AF_UNIX)
        except ConnectionRefusedError as e:
            raise UsbmuxConnectionError from e

    @staticmethod
    def create():
        # first attempt to connect with possibly the wrong version header (plist protocol)
        sock = MuxConnection.create_usbmux_socket()

        message = usbmuxd_request.build({
            'header': {'version': usbmuxd_version.PLIST, 'message': usbmuxd_msgtype.PLIST, 'tag': 1},
            'data': plistlib.dumps({'MessageType': 'ReadBUID'})
        })
        sock.send(message)
        response = usbmuxd_response.parse_stream(sock)

        # if we sent a bad request, we should re-create the socket in the correct version this time
        sock.close()
        sock = MuxConnection.create_usbmux_socket()

        if response.header.version == usbmuxd_version.BINARY:
            return BinaryMuxConnection(sock)
        elif response.header.version == usbmuxd_version.PLIST:
            return PlistMuxConnection(sock)

        raise MuxVersionError(f'usbmuxd returned unsupported version: {response.version}')

    def __init__(self, sock: SafeStreamSocket):
        self._sock = sock

        # after initiating the "Connect" packet, this same socket will be used to transfer data into the service
        # residing inside the target device. when this happens, we can no longer send/receive control commands to
        # usbmux on same socket
        self._connected = False

        # message sequence number. used when verifying the response matched the request
        self._tag = 1

        self.devices = []

    @abc.abstractmethod
    def _connect(self, device_id: int, port: int):
        """ initiate a "Connect" request to target port """
        pass

    @abc.abstractmethod
    def get_device_list(self, timeout: float = None):
        """
        request an update to current device list
        """
        pass

    def connect(self, device: MuxDevice, port: int) -> socket.socket:
        """ connect to a relay port on target machine and get a raw python socket object for the connection """
        self._connect(device.devid, socket.htons(port))
        self._connected = True
        return self._sock.sock

    def close(self):
        """ close current socket """
        self._sock.close()

    def _assert_not_connected(self):
        """ verify active state is in state for control messages """
        if self._connected:
            raise MuxException('Mux is connected, cannot issue control packets')


class BinaryMuxConnection(MuxConnection):
    """ old binary protocol """

    def __init__(self, sock: SafeStreamSocket):
        super().__init__(sock)
        self._version = usbmuxd_version.BINARY

    def get_device_list(self, timeout: float = None):
        """ use timeout to wait for the device list to be fully populated """
        self._assert_not_connected()
        end = time.time() + timeout
        self.listen()
        while time.time() < end:
            self._sock.settimeout(end - time.time())
            try:
                self._receive_device_state_update()
            except (BlockingIOError, StreamError):
                continue
            except IOError:
                try:
                    self._sock.setblocking(True)
                    self.close()
                except OSError:
                    pass
                raise MuxException('Exception in listener socket')

    def listen(self):
        """ start listening for events of attached and detached devices """
        self._send_receive(usbmuxd_msgtype.LISTEN)

    def _connect(self, device_id: int, port: int):
        self._send({'header': {'version': self._version,
                               'message': usbmuxd_msgtype.CONNECT,
                               'tag': self._tag},
                    'data': {'device_id': device_id, 'port': port},
                    })
        response = self._receive()
        if response.header.message != usbmuxd_msgtype.RESULT:
            raise MuxException(f'unexpected message type received: {response}')

        if response.data.result != usbmuxd_result.OK:
            raise MuxException(f'failed to connect to device: {device_id} at port: {port}. reason: '
                               f'{response.data.result}')

    def _send(self, data: Mapping):
        self._assert_not_connected()
        self._sock.send(usbmuxd_request.build(data))
        self._tag += 1

    def _receive(self, expected_tag: int = None):
        self._assert_not_connected()
        response = usbmuxd_response.parse_stream(self._sock)
        if expected_tag and response.header.tag != expected_tag:
            raise MuxException(f'Reply tag mismatch: expected {expected_tag}, got {response.header.tag}')
        return response

    def _send_receive(self, message_type: int):
        self._send({'header': {'version': self._version, 'message': message_type, 'tag': self._tag},
                    'data': b''})
        response = self._receive(self._tag - 1)
        if response.header.message != usbmuxd_msgtype.RESULT:
            raise MuxException(f'unexpected message type received: {response}')

        result = response.data.result
        if result != usbmuxd_result.OK:
            raise MuxException(f'{message_type} failed: error {result}')

    def _add_device(self, device: MuxDevice):
        self.devices.append(device)

    def _remove_device(self, device_id: int):
        self.devices = [device for device in self.devices if device.devid != device_id]

    def _receive_device_state_update(self):
        response = self._receive()
        if response.header.message == usbmuxd_msgtype.ADD:
            # old protocol only supported USB devices
            self._add_device(MuxDevice(response.data.device_id, response.data.serial_number, 'USB'))
        elif response.header.message == usbmuxd_msgtype.REMOVE:
            self._remove_device(response.data.device_id)
        else:
            raise MuxException(f'Invalid packet type received: {response}')


class PlistMuxConnection(BinaryMuxConnection):
    def __init__(self, sock: SafeStreamSocket):
        super().__init__(sock)
        self._version = usbmuxd_version.PLIST

    def listen(self) -> None:
        self._send_receive({'MessageType': 'Listen'})

    def get_pair_record(self, serial: str) -> Mapping:
        # serials are saved inside usbmuxd without '-'
        self._send({'MessageType': 'ReadPairRecord', 'PairRecordID': serial})
        response = self._receive(self._tag - 1)
        pair_record = response.get('PairRecordData')
        if pair_record is None:
            raise NotPairedError('device should be paired first')
        return plistlib.loads(pair_record)

    def get_device_list(self, timeout: float = None) -> None:
        """ get device list synchronously without waiting the timeout """
        self.devices = []
        self._send({'MessageType': 'ListDevices'})
        for response in self._receive(self._tag - 1)['DeviceList']:
            if response['MessageType'] == 'Attached':
                super()._add_device(MuxDevice(response['DeviceID'], response['Properties']['SerialNumber'],
                                              response['Properties']['ConnectionType']))
            elif response['MessageType'] == 'Detached':
                super()._remove_device(response['DeviceID'])
            else:
                raise MuxException(f'Invalid packet type received: {response}')

    def get_buid(self) -> str:
        """ get SystemBUID """
        self._send({'MessageType': 'ReadBUID'})
        return self._receive(self._tag - 1)['BUID']

    def save_pair_record(self, serial: str, device_id: int, record_data: bytes):
        # serials are saved inside usbmuxd without '-'
        self._send_receive({'MessageType': 'SavePairRecord',
                            'PairRecordID': serial,
                            'PairRecordData': record_data,
                            'DeviceID': device_id})

    def _connect(self, device_id: int, port: int):
        self._send_receive({'MessageType': 'Connect', 'DeviceID': device_id, 'PortNumber': port})

    def _send(self, data: Mapping):
        request = {'ClientVersionString': 'qt4i-usbmuxd', 'ProgName': 'pymobiledevice3', 'kLibUSBMuxVersion': 3}
        request.update(data)
        super()._send({'header': {'version': self._version,
                                  'message': usbmuxd_msgtype.PLIST,
                                  'tag': self._tag},
                       'data': plistlib.dumps(request),
                       })

    def _receive(self, expected_tag: int = None) -> Mapping:
        response = super()._receive(expected_tag=expected_tag)
        if response.header.message != usbmuxd_msgtype.PLIST:
            raise MuxException(f'Received non-plist type {response}')
        return plistlib.loads(response.data)

    def _send_receive(self, data: Mapping):
        self._send(data)
        response = self._receive(self._tag - 1)
        if response['MessageType'] != 'Result':
            raise MuxException(f'got an invalid message: {response}')
        if response['Number'] != 0:
            raise MuxException(f'got an error message: {response}')


def create_mux() -> MuxConnection:
    return MuxConnection.create()


def list_devices() -> List[MuxDevice]:
    mux = create_mux()
    mux.get_device_list(0.1)
    devices = mux.devices
    mux.close()
    return devices


def select_device(udid: str = None, connection_type: str = None) -> Optional[MuxDevice]:
    """
    select a UsbMux device according to given arguments.
    if more than one device could be selected, always prefer the usb one.
    """
    tmp = None
    for device in list_devices():
        if connection_type is not None and device.connection_type != connection_type:
            # if a specific connection_type was desired and not of this one then skip
            continue

        if udid is not None and not device.matches_udid(udid):
            # if a specific udid was desired and not of this one then skip
            continue

        # save best result as a temporary
        tmp = device

        if device.is_usb:
            # always prefer usb connection
            return device

    return tmp


def select_devices_by_connection_type(connection_type: str) -> List[MuxDevice]:
    """
    select all UsbMux devices by connection type
    """
    tmp = []
    for device in list_devices():
        if device.connection_type == connection_type:
            tmp.append(device)

    return tmp



import asyncio

import click
import IPython
import uvicorn
from pygments import highlight, lexers, formatters

from pymobiledevice3.cli.cli_common import Command, wait_return
from pymobiledevice3.exceptions import WirError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.webinspector import WebinspectorService, SAFARI
from pymobiledevice3.services.web_protocol.driver import WebDriver, Cookie, By
from pymobiledevice3.services.web_protocol.cdp_server import app


@click.group()
def cli():
    """ webinspector cli """
    pass


@cli.group()
def webinspector():
    """ webinspector options """
    pass


@webinspector.command(cls=Command)
@click.option('-v', '--verbose', is_flag=True)
def opened_tabs(lockdown: LockdownClient, verbose):
    """
    Show All opened tabs.
    Opt in:

        Settings -> Safari -> Advanced -> Web Inspector
    """
    inspector = WebinspectorService(lockdown=lockdown, loop=asyncio.get_event_loop())
    inspector.connect()
    while not inspector.connected_application:
        inspector.flush_input()
    inspector.get_open_pages()
    # Best effort.
    inspector.flush_input(2)
    for app_id, app_ in inspector.connected_application.items():
        if app_id not in inspector.application_pages:
            continue
        if verbose:
            print(f'{app_.name}    id: {app_id}')
        else:
            print(app_.name)
        for page_id, page in inspector.application_pages[app_id].items():
            if verbose:
                print(f' - {page.web_url}    id: {page_id}')
            else:
                print(f' - {page.web_url}')
    inspector.close()


@webinspector.command(cls=Command)
@click.argument('url')
def launch(lockdown: LockdownClient, url):
    """
    Open a specific URL in Safari.
    Opt in:

        Settings -> Safari -> Advanced -> Web Inspector

        Settings -> Safari -> Advanced -> Remote Automation
    """
    inspector = WebinspectorService(lockdown=lockdown)
    inspector.connect()
    safari = inspector.open_app(SAFARI)
    session = inspector.automation_session(safari)
    driver = WebDriver(session)
    print('Starting session')
    driver.start_session()
    print('Getting URL')
    driver.get(url)
    wait_return()
    session.stop_session()
    inspector.close()


SHELL_USAGE = '''
# This shell allows you to control the web with selenium like API.
# The first thing you should do is creating a session:
driver.start_session()

# Then, you can navigate by using the get method:
driver.get("https://google.com")

# You can search for a specific element using the `By` enum:
driver.find_element(By.TAG_NAME, 'input')

# You can also add cookies:
driver.add_cookie(
    Cookie(name='tz', value='UTC', domain='.github.com', path='/', expires=0, httpOnly=False, secure=True,
    session=True, sameSite='None')
)

# See selenium api for more features.
'''


@webinspector.command(cls=Command)
def shell(lockdown: LockdownClient):
    """
    Opt in:

        Settings -> Safari -> Advanced -> Web Inspector

        Settings -> Safari -> Advanced -> Remote Automation
    """
    inspector = WebinspectorService(lockdown=lockdown)
    inspector.connect()
    safari = inspector.open_app(SAFARI)
    session = inspector.automation_session(safari)
    driver = WebDriver(session)
    try:
        IPython.embed(
            header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.TerminalTrueColorFormatter(style='native')),
            user_ns={
                'driver': driver,
                'Cookie': Cookie,
                'By': By,
            })
    finally:
        session.stop_session()
        inspector.close()


@webinspector.command(cls=Command)
@click.argument('url', required=False, default='')
def jsshell(lockdown: LockdownClient, url):
    """
    Opt in:

        Settings -> Safari -> Advanced -> Web Inspector

        Settings -> Safari -> Advanced -> Remote Automation
    """
    inspector = WebinspectorService(lockdown=lockdown)
    inspector.connect()
    safari = inspector.open_app(SAFARI)
    session = inspector.automation_session(safari)
    driver = WebDriver(session)
    try:
        driver.start_session()
        if url:
            driver.get(url)
        while True:
            exp = input('> ')
            try:
                print(driver.execute_script(f'return {exp}'))
            except WirError as e:
                print(e)
    finally:
        session.stop_session()
        inspector.close()


udid = ''


def create_app():
    inspector = WebinspectorService(lockdown=LockdownClient(udid))
    app.state.inspector = inspector
    return app


@webinspector.command(cls=Command)
@click.option('--host', default='127.0.0.1')
@click.option('--port', type=click.INT, default=9222)
def cdp(lockdown: LockdownClient, host, port):
    global udid
    udid = lockdown.udid
    uvicorn.run('pymobiledevice3.cli.webinspector:create_app', host=host, port=port, factory=True,
                ws_ping_timeout=None, ws='wsproto', loop='asyncio')



import asn1
import logging
import struct
from typing import Optional

logger = logging.getLogger(__name__)


def img4_get_component_tag(compname):
    component_tags = {
        'ACIBT': b'acib',
        'ACIBTLPEM': b'lpbt',
        'ACIWIFI': b'aciw',
        'ANE': b'anef',
        'ANS': b'ansf',
        'AOP': b'aopf',
        'AVE': b'avef',
        'Alamo': b'almo',
        'Ap,ANE1': b'ane1',
        'Ap,ANE2': b'ane2',
        'Ap,ANE3': b'ane3',
        'Ap,AudioAccessibilityBootChime': b'auac',
        'Ap,AudioBootChime': b'aubt',
        'Ap,AudioPowerAttachChime': b'aupr',
        'Ap,BootabilityBrainTrustCache': b'trbb',
        'Ap,CIO': b'ciof',
        'Ap,HapticAssets': b'hpas',
        'Ap,LocalBoot': b'lobo',
        'Ap,LocalPolicy': b'lpol',
        'Ap,NextStageIM4MHash': b'nsih',
        'Ap,RecoveryOSPolicyNonceHash': b'ronh',
        'Ap,RestoreANE1': b'ran1',
        'Ap,RestoreANE2': b'ran2',
        'Ap,RestoreANE3': b'ran3',
        'Ap,RestoreCIO': b'rcio',
        'Ap,RestoreTMU': b'rtmu',
        'Ap,Scorpius': b'scpf',
        'Ap,SystemVolumeCanonicalMetadata': b'msys',
        'Ap,TMU': b'tmuf',
        'Ap,VolumeUUID': b'vuid',
        'Ap,rOSLogo1': b'rlg1',
        'Ap,rOSLogo2': b'rlg2',
        'AppleLogo': b'logo',
        'AudioCodecFirmware': b'acfw',
        'BatteryCharging': b'glyC',
        'BatteryCharging0': b'chg0',
        'BatteryCharging1': b'chg1',
        'BatteryFull': b'batF',
        'BatteryLow0': b'bat0',
        'BatteryLow1': b'bat1',
        'BatteryPlugin': b'glyP',
        'CFELoader': b'cfel',
        'CrownFirmware': b'crwn',
        'DCP': b'dcpf',
        'Dali': b'dali',
        'DeviceTree': b'dtre',
        'Diags': b'diag',
        'EngineeringTrustCache': b'dtrs',
        'ExtDCP': b'edcp',
        'GFX': b'gfxf',
        'Hamm': b'hamf',
        'Homer': b'homr',
        'ISP': b'ispf',
        'InputDevice': b'ipdf',
        'KernelCache': b'krnl',
        'LLB': b'illb',
        'LeapHaptics': b'lphp',
        'Liquid': b'liqd',
        'LoadableTrustCache': b'ltrs',
        'LowPowerWallet0': b'lpw0',
        'LowPowerWallet1': b'lpw1',
        'LowPowerWallet2': b'lpw2',
        'MacEFI': b'mefi',
        'MtpFirmware': b'mtpf',
        'Multitouch': b'mtfw',
        'NeedService': b'nsrv',
        'OS': b'OS\0\0',
        'OSRamdisk': b'osrd',
        'PEHammer': b'hmmr',
        'PERTOS': b'pert',
        'PHLEET': b'phlt',
        'PMP': b'pmpf',
        'PersonalizedDMG': b'pdmg',
        'RBM': b'rmbt',
        'RTP': b'rtpf',
        'Rap,SoftwareBinaryDsp1': b'sbd1',
        'Rap,RTKitOS': b'rkos',
        'Rap,RestoreRTKitOS': b'rrko',
        'RecoveryMode': b'recm',
        'RestoreANS': b'rans',
        'RestoreDCP': b'rdcp',
        'RestoreDeviceTree': b'rdtr',
        'RestoreExtDCP': b'recp',
        'RestoreKernelCache': b'rkrn',
        'RestoreLogo': b'rlgo',
        'RestoreRTP': b'rrtp',
        'RestoreRamDisk': b'rdsk',
        'RestoreSEP': b'rsep',
        'RestoreTrustCache': b'rtsc',
        'SCE': b'scef',
        'SCE1Firmware': b'sc1f',
        'SEP': b'sepi',
        'SIO': b'siof',
        'StaticTrustCache': b'trst',
        'SystemLocker': b'lckr',
        'SystemVolume': b'isys',
        'WCHFirmwareUpdater': b'wchf',
        'ftap': b'ftap',
        'ftsp': b'ftsp',
        'iBEC': b'ibec',
        'iBSS': b'ibss',
        'iBoot': b'ibot',
        'iBootData': b'ibdt',
        'iBootDataStage1': b'ibd1',
        'iBootTest': b'itst',
        'rfta': b'rfta',
        'rfts': b'rfts',
    }

    return component_tags.get(compname)


def asn1_find_element(index: int, type_: int, data: bytes) -> Optional[int]:
    el_type = 0

    # verify data integrity
    off = 0
    if data[off] != (asn1.Types.Constructed | asn1.Numbers.Sequence):
        return None
    off += 1

    # check data size
    offsets = {0x84: 4, 0x83: 3, 0x82: 2, 0x81: 1}
    off += 1 + offsets.get(data[off], 0)

    # find the element we are searching
    for i in range(0, index + 1):
        el_type = data[off]
        off += 1
        el_size = data[off]
        off += 1

        if i == index:
            break

        off += el_size

    # check element type
    if el_type != type_:
        return None

    return off


def stitch_component(name: str, data: bytes, tss):
    logger.info(f'Personalizing IMG4 component {name}...')

    blob = tss.ap_img4_ticket

    # first we need check if we have to change the tag for the given component
    tag_off = asn1_find_element(1, asn1.Numbers.IA5String, data)

    if tag_off is not None:
        logger.debug('Tag found')
        component_name_tag = {
            'RestoreKernelCache': b'rkrn',
            'RestoreDeviceTree': b'rdtr',
            'RestoreSEP': b'rsep',
            'RestoreLogo': b'rlgo',
            'RestoreTrustCache': b'rtsc',
            'RestoreDCP': b'rdcp',
            'Ap,RestoreTMU': b'rtmu',
            'Ap,RestoreCIO': b'rcio',
            'Ap,DCP2': b'dcp2',
        }
        if name in component_name_tag:
            first_len = len(data)

            data = bytearray(data)
            chunk = component_name_tag[name]
            data[tag_off:tag_off + len(chunk)] = chunk
            data = bytes(data)

            assert len(data) == first_len
    else:
        logging.warning('not asn1 sequence')

    # create element header for the "IMG4" magic
    encoder = asn1.Encoder()
    encoder.start()

    encoder.enter(asn1.Numbers.Sequence)

    # create element header for the "IMG4" magic
    encoder.write(b'IMG4', asn1.Numbers.IA5String)

    decoder = asn1.Decoder()
    decoder.start(data)
    encoder.write(decoder.read()[1], nr=asn1.Numbers.Sequence, typ=asn1.Types.Constructed)

    encoder.enter(0, cls=asn1.Classes.Context)

    decoder = asn1.Decoder()
    decoder.start(blob)
    encoder.write(decoder.read()[1], nr=asn1.Numbers.Sequence, typ=asn1.Types.Constructed)

    encoder.leave()

    # hack if we have a *-TBM entry for the give
    tbm_dict = tss.get(f'{name}-TBM')
    if tbm_dict is not None:
        # now construct IM4R
        encoder.enter(asn1.Numbers.Boolean, cls=asn1.Classes.Context)
        encoder.enter(asn1.Numbers.Sequence)
        encoder.write(b'IM4R', asn1.Numbers.IA5String)
        encoder.enter(asn1.Numbers.Set)

        for tbm_component in (b'ucon', b'ucer'):
            # write priv ucon element
            encoder.enter(struct.unpack('>I', tbm_component)[0], cls=asn1.Classes.Private)

            # write ucon IA5STRING and ucon data
            encoder.enter(asn1.Numbers.Sequence)
            encoder.write(tbm_component, asn1.Numbers.IA5String)
            encoder.write(tbm_dict[tbm_component.decode()])
            encoder.leave()

            encoder.leave()

        encoder.leave()
        encoder.leave()
        encoder.leave()

    encoder.leave()

    return encoder.output()



from pymobiledevice3.exceptions import WirError
from pymobiledevice3.services.web_protocol.automation_session import RESOURCES, Point, Size, Rect, MouseButton, \
    MouseInteraction, VIRTUAL_KEYS, KeyboardInteractionType, MODIFIER_TO_KEY
from pymobiledevice3.services.web_protocol.selenium_api import By, SeleniumApi

IS_EDITABLE = (RESOURCES / 'is_editable.js').read_text()
ELEMENT_CLEAR = (RESOURCES / 'element_clear.js').read_text()
GET_ATTRIBUTE = (RESOURCES / 'get_attribute.js').read_text()
ELEMENT_ATTRIBUTE = (RESOURCES / 'element_attribute.js').read_text()
IS_DISPLAYED = (RESOURCES / 'is_displayed.js').read_text()
IS_ENABLED = (RESOURCES / 'is_enabled.js').read_text()
FOCUS = (RESOURCES / 'focus.js').read_text()


class WebElement(SeleniumApi):
    def __init__(self, session, id_):
        """
        :param pymobiledevice3.services.web_protocol.automation_session.AutomationSession session: Automation session.
        :param dict id_: Element id.
        """
        self.session = session
        self.id_ = id_
        self.node_id = id_[f'session-node-{self.session.id_}']

    def clear(self):
        """ Clears the text if it's a text entry element. """
        if not self.is_editable():
            return
        rect, center, is_obscured = self._compute_layout()
        if rect is None or center is None:
            return
        self._evaluate_js_function(ELEMENT_CLEAR)

    def click(self):
        """ Clicks the element. """
        rect, center, is_obscured = self._compute_layout(use_viewport=True)
        if rect is None or is_obscured or center is None:
            return
        if self.tag_name == 'option':
            self._select_option_element()
        else:
            self.session.perform_mouse_interaction(center.x, center.y, MouseButton.LEFT, MouseInteraction.SINGLE_CLICK)

    def find_element(self, by=By.ID, value=None):
        """ Find an element given a By strategy and locator. """
        elem = self.session.find_elements(by, value, root=self.id_)
        return None if elem is None else WebElement(self.session, elem)

    def find_elements(self, by=By.ID, value=None):
        """ Find elements given a By strategy and locator. """
        elements = self.session.find_elements(by, value, single=False, root=self.id_)
        return list(map(lambda elem: WebElement(self.session, elem), elements))

    def get_attribute(self, name: str) -> str:
        """ Gets the given attribute or property of the element. """
        return self.session.execute_script(f'return ({GET_ATTRIBUTE}).apply(null, arguments);', self.id_, name)

    def get_dom_attribute(self, name: str) -> str:
        """ Gets the given attribute of the element. """
        return self._evaluate_js_function(ELEMENT_ATTRIBUTE, name)

    def get_property(self, name: str) -> str:
        """ Gets the given property of the element. """
        return self._evaluate_js_function(f'function(element) {{ return element.{name}; }}')

    def is_displayed(self) -> bool:
        """ Whether the element is visible to a user. """
        return self.session.execute_script(f'return ({IS_DISPLAYED}).apply(null, arguments);', self.id_)

    def is_enabled(self) -> bool:
        """ Returns whether the element is enabled. """
        return self._evaluate_js_function(IS_ENABLED)

    def is_selected(self) -> bool:
        """ Returns whether the element is selected. Can be used to check if a checkbox or radio button is selected. """
        return bool(self.get_dom_attribute('selected'))

    @property
    def location(self) -> Point:
        """ The location of the element in the renderable canvas. """
        rect = self.rect
        return Point(x=rect.x, y=rect.y)

    @property
    def location_once_scrolled_into_view(self) -> Point:
        """ Returns the top lefthand corner location on the screen, or ``None`` if the element is not visible. """
        rect = self.session.execute_script(
            'arguments[0].scrollIntoView(true); return arguments[0].getBoundingClientRect(); ', self.id_
        )
        return Point(x=round(rect['x']), y=round(rect['y']))

    @property
    def rect(self) -> Rect:
        """ The size and location of the element. """
        return self._compute_layout(scroll_if_needed=False)[0]

    @property
    def screenshot_as_base64(self) -> str:
        """ Gets the screenshot of the current element as a base64 encoded string. """
        return self.session.screenshot_as_base64(scroll=True, node_id=self.node_id)

    def send_keys(self, value):
        """
        Simulates typing into the element.
        :param value: A string for typing, or setting form fields.
        """
        self._evaluate_js_function(FOCUS)
        interactions = []
        sticky_modifier = set()
        for key in value:
            if key in VIRTUAL_KEYS:
                virtual_key, modifier = VIRTUAL_KEYS[key]
                interaction = {'type': KeyboardInteractionType.INSERT_BY_KEY, 'key': virtual_key}
                if modifier is not None:
                    sticky_modifier ^= {modifier}
                    interaction['type'] = (KeyboardInteractionType.KEY_PRESS if modifier in sticky_modifier
                                           else KeyboardInteractionType.KEY_RELEASE)
                interactions.append(interaction)
            else:
                interactions.append({'type': KeyboardInteractionType.INSERT_BY_KEY, 'text': key})
        for modifier in sticky_modifier:
            interactions.append({'type': KeyboardInteractionType.KEY_RELEASE, 'key': MODIFIER_TO_KEY[modifier]})
        self.session.perform_keyboard_interactions(interactions)

    @property
    def size(self) -> Size:
        """ The size of the element. """
        rect = self.rect
        return Size(height=rect.height, width=rect.width)

    def submit(self):
        """ Submits a form. """
        form = self.find_element(By.XPATH, "./ancestor-or-self::form")
        submit_code = (
            'var e = arguments[0].ownerDocument.createEvent(\'Event\');'
            'e.initEvent(\'submit\', true, true);'
            'if (arguments[0].dispatchEvent(e)) { arguments[0].submit() }'
        )
        self.session.execute_script(submit_code, form.id_)

    @property
    def tag_name(self) -> str:
        """ This element's ``tagName`` property. """
        return self._evaluate_js_function('function(element) { return element.tagName.toLowerCase() }')

    @property
    def text(self) -> str:
        """ The text of the element. """
        return self._evaluate_js_function(
            'function(element) { return element.innerText.replace(/^[^\\S\\xa0]+|[^\\S\\xa0]+$/g, "") }'
        )

    def touch(self):
        """ Simulate touch interaction on the element. """
        rect, center, is_obscured = self._compute_layout(use_viewport=True)
        self.session.perform_interaction_sequence(
            [{'sourceId': self.session.id_, 'sourceType': 'Touch'}],
            [{'states': [
                {'sourceId': self.session.id_, 'location': {'x': center.x, 'y': center.y}}
            ]}]
        )

    def value_of_css_property(self, property_name) -> str:
        """ The value of a CSS property. """
        return self._evaluate_js_function(
            'function(element) {'
            f' return document.defaultView.getComputedStyle(element).getPropertyValue("{property_name}"); '
            '}'
        )

    def is_editable(self) -> bool:
        """ Returns whether the element is editable. """
        return self._evaluate_js_function(IS_EDITABLE)

    def _compute_layout(self, scroll_if_needed=True, use_viewport=False):
        try:
            result = self.session.compute_element_layout(self.node_id, scroll_if_needed,
                                                         'LayoutViewport' if use_viewport else 'Page')
        except WirError:
            return

        origin = result['rect']['origin']
        size = result['rect']['size']
        rect = Rect(x=round(origin['x']), y=round(origin['y']), width=size['width'], height=size['height'])
        center = Point(x=result['inViewCenterPoint']['x'], y=result['inViewCenterPoint']['y'])
        return rect, center, result['isObscured']

    def _select_option_element(self):
        self.session.select_option_element(self.node_id)

    def _evaluate_js_function(self, function, *args):
        return self.session.evaluate_js_function(function, self.id_, *args)

    def __eq__(self, other):
        return self.id_ == other.id_



import binascii
import hashlib
import logging
import os
import plistlib
import struct
import tempfile
import traceback
import zipfile
from io import BytesIO
from typing import Optional, Mapping

from tqdm import tqdm, trange

from pymobiledevice3.exceptions import PyMobileDevice3Exception, NoDeviceConnectedError, ConnectionFailedError
from pymobiledevice3.restore.asr import ASRClient
from pymobiledevice3.restore.base_restore import BaseRestore, RESTORE_VARIANT_MACOS_RECOVERY_OS, \
    RESTORE_VARIANT_ERASE_INSTALL, RESTORE_VARIANT_UPGRADE_INSTALL
from pymobiledevice3.restore.consts import lpol_file, PROGRESS_BAR_OPERATIONS
from pymobiledevice3.restore.device import Device
from pymobiledevice3.restore.fdr import start_fdr_thread, fdr_type, FDRClient
from pymobiledevice3.restore.ftab import Ftab
from pymobiledevice3.restore.recovery import Recovery, Behavior
from pymobiledevice3.restore.restore_options import RestoreOptions
from pymobiledevice3.restore.restored_client import RestoredClient
from pymobiledevice3.restore.tss import TSSRequest, TSSResponse
from pymobiledevice3.service_connection import ServiceConnection
from pymobiledevice3.utils import plist_access_path

known_errors = {
    0xFFFFFFFFFFFFFFFF: 'verification error',
    6: 'disk failure',
    14: 'fail',
    27: 'failed to mount filesystems',
    51: 'failed to load SEP firmware',
    53: 'failed to recover FDR data',
    1015: 'X-Gold Baseband Update Failed. Defective Unit?',
}


class Restore(BaseRestore):
    def __init__(self, ipsw: BytesIO, device: Device, tss=None, behavior: Behavior = Behavior.Update, ignore_fdr=False):
        super().__init__(ipsw, device, tss, behavior, logger=logging.getLogger(__name__))
        self.recovery = Recovery(ipsw, device, tss=tss, behavior=behavior)
        self.bbtss = None  # type: Optional[TSSResponse]
        self._restored = None  # type: Optional[RestoredClient]
        self._restore_finished = False

        # used when ignore_fdr=True, to store an active FDR connection just to make the device believe it can actually
        # perform an FDR communication, but without really establishing any
        self._fdr = None  # type: Optional[ServiceConnection]
        self._ignore_fdr = ignore_fdr

        # query preflight info while device may still be in normal mode
        self._preflight_info = self.device.preflight_info

        # prepare progress bar for OS component verify
        self._pb_verify_restore = None
        self._pb_verify_restore_old_value = None

        self._handlers = {
            # data request messages are sent by restored whenever it requires
            # files sent to the server by the client. these data requests include
            # SystemImageData, RootTicket, KernelCache, NORData and BasebandData requests
            'DataRequestMsg': self.handle_data_request_msg,

            # restore logs are available if a previous restore failed
            'PreviousRestoreLogMsg': self.handle_previous_restore_log_msg,

            # progress notification messages sent by the restored inform the client
            # of it's current operation and sometimes percent of progress is complete
            'ProgressMsg': self.handle_progress_msg,

            # status messages usually indicate the current state of the restored
            # process or often to signal an error has been encountered
            'StatusMsg': self.handle_status_msg,

            # checkpoint notifications
            'CheckpointMsg': self.handle_checkpoint_msg,

            # baseband update message
            'BBUpdateStatusMsg': self.handle_bb_update_status_msg,

            # baseband updater output data request
            'BasebandUpdaterOutputData': self.handle_baseband_updater_output_data,
        }

        self._data_request_handlers = {
            # this request is sent when restored is ready to receive the filesystem
            'SystemImageData': self.send_filesystem,

            'BuildIdentityDict': self.send_buildidentity,
            'PersonalizedBootObjectV3': self.send_personalized_boot_object_v3,
            'SourceBootObjectV4': self.send_source_boot_object_v4,
            'RecoveryOSLocalPolicy': self.send_restore_local_policy,

            # this request is sent when restored is ready to receive the filesystem
            'RecoveryOSASRImage': self.send_filesystem,

            # Send RecoveryOS RTD
            'RecoveryOSRootTicketData': self.send_recovery_os_root_ticket,

            # send RootTicket (== APTicket from the TSS request)
            'RootTicket': self.send_root_ticket,

            'NORData': self.send_nor,
            'BasebandData': self.send_baseband_data,
            'FDRTrustData': self.send_fdr_trust_data,
            'FirmwareUpdaterData': self.send_firmware_updater_data,

            # TODO: verify
            'FirmwareUpdaterPreflight': self.send_firmware_updater_preflight,
        }

        self._data_request_components = {
            'KernelCache': self.send_component,
            'DeviceTree': self.send_component,
        }

    def send_filesystem(self, message: Mapping):
        self.logger.info('about to send filesystem...')

        self.logger.info('connecting to ASR...')
        while True:
            try:
                asr = ASRClient(self._restored.udid)
                break
            except ConnectionFailedError:
                pass

        self.logger.info('connected to ASR')

        # this step sends requested chunks of data from various offsets to asr, so
        # it can validate the filesystem before installing it
        self.logger.info('validating the filesystem')

        with self.ipsw.open_path(self.build_identity.get_component_path('OS')) as filesystem:
            asr.perform_validation(filesystem)
            self.logger.info('filesystem validated')

            # once the target filesystem has been validated, ASR then requests the
            # entire filesystem to be sent.
            self.logger.info('sending filesystem now...')
            asr.send_payload(filesystem)

    def get_build_identity_from_request(self, msg):
        return self.get_build_identity(msg['Arguments'].get('IsRecoveryOS', False))

    def send_buildidentity(self, message: Mapping):
        self.logger.info('About to send BuildIdentity Dict...')
        req = {'BuildIdentityDict': dict(self.get_build_identity_from_request(message))}
        arguments = message['Arguments']
        variant = arguments.get('Variant')

        if variant:
            req['Variant'] = variant
        else:
            req['Variant'] = 'Erase'

        self.logger.info('Sending BuildIdentityDict now...')
        self._restored.send(req)

    def extract_global_manifest(self):
        build_info = self.build_identity.get('Info')
        if build_info is None:
            raise PyMobileDevice3Exception('build identity does not contain an "Info" element')

        device_class = build_info.get('DeviceClass')
        if device_class is None:
            raise PyMobileDevice3Exception('build identity does not contain an "DeviceClass" element')

        macos_variant = build_info.get('MacOSVariant')
        if macos_variant is None:
            raise PyMobileDevice3Exception('build identity does not contain an "MacOSVariant" element')

        # The path of the global manifest is hardcoded. There's no pointer to in the build manifest.
        return self.ipsw.get_global_manifest(macos_variant, device_class)

    def send_personalized_boot_object_v3(self, message: Mapping):
        self.logger.debug('send_personalized_boot_object_v3')
        image_name = message['Arguments']['ImageName']
        component_name = image_name
        self.logger.info(f'About to send {component_name}...')

        if image_name == '__GlobalManifest__':
            data = self.extract_global_manifest()
        elif image_name == '__RestoreVersion__':
            data = self.ipsw.restore_version
        elif image_name == '__SystemVersion__':
            data = self.ipsw.system_version
        else:
            data = self.build_identity.get_component(component_name, tss=self.recovery.tss).personalized_data

        self.logger.info(f'Sending {component_name} now...')
        chunk_size = 8192
        for i in trange(0, len(data), chunk_size):
            self._restored.send({'FileData': data[i:i + chunk_size]})

        # Send FileDataDone
        self._restored.send({'FileDataDone': True})

        self.logger.info(f'Done sending {component_name}')

    def send_source_boot_object_v4(self, message: Mapping):
        self.logger.debug('send_source_boot_object_v4')
        image_name = message['Arguments']['ImageName']
        component_name = image_name
        self.logger.info(f'About to send {component_name}...')

        if image_name == '__GlobalManifest__':
            data = self.extract_global_manifest()
        elif image_name == '__RestoreVersion__':
            data = self.ipsw.restore_version
        elif image_name == '__SystemVersion__':
            data = self.ipsw.system_version
        else:
            data = self.get_build_identity_from_request(message) \
                .get_component(component_name, tss=self.recovery.tss).data

        self.logger.info(f'Sending {component_name} now...')
        chunk_size = 8192
        for i in trange(0, len(data), chunk_size):
            self._restored.send({'FileData': data[i:i + chunk_size]})

        # Send FileDataDone
        self._restored.send({'FileDataDone': True})

        self.logger.info(f'Done sending {component_name}')

    def get_recovery_os_local_policy_tss_response(self, args, build_identity=None):
        if build_identity is None:
            build_identity = self.build_identity

        # populate parameters
        parameters = {
            'ApECID': self.device.ecid,
            'Ap,LocalBoot': True,
            'ApProductionMode': True,
            'ApSecurityMode': True,
            'ApSupportsImg4': True,
        }

        build_identity.populate_tss_request_parameters(parameters)

        # Add Ap,LocalPolicy
        lpol = {
            'Digest': hashlib.sha384(lpol_file).digest(),
            'Trusted': True,
        }

        parameters['Ap,LocalPolicy'] = lpol

        parameters['Ap,NextStageIM4MHash'] = args['Ap,NextStageIM4MHash']
        parameters['Ap,RecoveryOSPolicyNonceHash'] = args['Ap,RecoveryOSPolicyNonceHash']

        vol_uuid = args['Ap,VolumeUUID']
        vol_uuid = binascii.unhexlify(vol_uuid.replace('-', ''))
        parameters['Ap,VolumeUUID'] = vol_uuid

        # create basic request
        request = TSSRequest()

        # add common tags from manifest
        request.add_local_policy_tags(parameters)

        self.logger.info('Requesting SHSH blobs...')
        return request.send_receive()

    def get_build_identity(self, is_recovery_os: bool):
        if is_recovery_os:
            variant = RESTORE_VARIANT_MACOS_RECOVERY_OS
        elif self.build_identity.restore_behavior == Behavior.Erase:
            variant = RESTORE_VARIANT_ERASE_INSTALL
        else:
            variant = RESTORE_VARIANT_UPGRADE_INSTALL

        return self.ipsw.build_manifest.get_build_identity(self.device.hardware_model, variant=variant)

    def send_restore_local_policy(self, message: Mapping):
        component = 'Ap,LocalPolicy'

        # The Update mode does not have a specific build identity for the recovery os.
        build_identity = self.get_build_identity(self.build_identity.restore_behavior == Behavior.Erase)
        tss_localpolicy = self.get_recovery_os_local_policy_tss_response(message['Arguments'],
                                                                         build_identity=build_identity)

        self._restored.send({'Ap,LocalPolicy': build_identity.get_component(component, tss=tss_localpolicy,
                                                                            data=lpol_file).personalized_data})

    def send_recovery_os_root_ticket(self, message: Mapping):
        self.logger.info('About to send RecoveryOSRootTicket...')

        if self.recovery.tss_recoveryos_root_ticket is None:
            raise PyMobileDevice3Exception('Cannot send RootTicket without TSS')

        if self.device.is_image4_supported:
            data = self.recovery.tss_recoveryos_root_ticket.ap_img4_ticket
        else:
            data = self.recovery.tss.ap_ticket

        req = {}
        if data:
            req['RootTicketData'] = data
        else:
            self.logger.warning('not sending RootTicketData (no data present)')

        self.logger.info('Sending RecoveryOSRootTicket now...')
        self._restored.send(req)

    def send_root_ticket(self, message: Mapping):
        self.logger.info('About to send RootTicket...')

        if self.recovery.tss is None:
            raise PyMobileDevice3Exception('Cannot send RootTicket without TSS')

        self.logger.info('Sending RootTicket now...')
        self._restored.send({'RootTicketData': self.recovery.tss.ap_img4_ticket})

    def send_nor(self, message: Mapping):
        self.logger.info('About to send NORData...')
        flash_version_1 = False
        llb_path = self.build_identity.get_component('LLB', tss=self.recovery.tss).path
        llb_filename_offset = llb_path.find('LLB')

        arguments = message.get('Arguments')
        if arguments:
            flash_version_1 = arguments.get('FlashVersion1', False)

        if llb_filename_offset == -1:
            raise PyMobileDevice3Exception('Unable to extract firmware path from LLB filename')

        firmware_path = llb_path[:llb_filename_offset - 1]
        self.logger.info(f'Found firmware path: {firmware_path}')

        firmware_files = dict()
        try:
            firmware = self.ipsw.get_firmware(firmware_path)
            firmware_files = firmware.get_files()
        except KeyError:
            self.logger.info('Getting firmware manifest from build identity')
            build_id_manifest = self.build_identity['Manifest']
            for component, manifest_entry in build_id_manifest.items():
                if isinstance(manifest_entry, dict):
                    is_fw = plist_access_path(manifest_entry, ('Info', 'IsFirmwarePayload'), bool)
                    loaded_by_iboot = plist_access_path(manifest_entry, ('Info', 'IsLoadedByiBoot'), bool)
                    is_secondary_fw = plist_access_path(manifest_entry, ('Info', 'IsSecondaryFirmwarePayload'), bool)

                    if is_fw or (is_secondary_fw and loaded_by_iboot):
                        comp_path = plist_access_path(manifest_entry, ('Info', 'Path'))
                        if comp_path:
                            firmware_files[component] = comp_path

        if not firmware_files:
            raise PyMobileDevice3Exception('Unable to get list of firmware files.')

        component = 'LLB'
        llb_data = self.build_identity.get_component(component, tss=self.recovery.tss,
                                                     path=llb_path).personalized_data
        req = {'LlbImageData': llb_data}

        if flash_version_1:
            norimage = {}
        else:
            norimage = []

        for component, comppath in firmware_files.items():
            if component in ('LLB', 'RestoreSEP'):
                # skip LLB, it's already passed in LlbImageData
                # skip RestoreSEP, it's passed in RestoreSEPImageData
                continue

            nor_data = self.build_identity.get_component(component, tss=self.recovery.tss,
                                                         path=comppath).personalized_data

            if flash_version_1:
                norimage[component] = nor_data
            else:
                # make sure iBoot is the first entry in the array
                if component.startswith('iBoot'):
                    norimage = [nor_data] + norimage
                else:
                    norimage.append(nor_data)

        req['NorImageData'] = norimage

        for component in ('RestoreSEP', 'SEP'):
            comp = self.build_identity.get_component(component, tss=self.recovery.tss)
            if comp.path:
                req[f'{component}ImageData'] = comp.personalized_data

        self.logger.info('Sending NORData now...')
        self._restored.send(req)

    @staticmethod
    def get_bbfw_fn_for_element(elem):
        bbfw_fn_elem = {
            # ICE3 firmware files
            'RamPSI': 'psi_ram.fls',
            'FlashPSI': 'psi_flash.fls',
            # Trek firmware files
            'eDBL': 'dbl.mbn',
            'RestoreDBL': 'restoredbl.mbn',
            # Phoenix/Mav4 firmware files
            'DBL': 'dbl.mbn',
            'ENANDPRG': 'ENPRG.mbn',
            # Mav5 firmware files
            'RestoreSBL1': 'restoresbl1.mbn',
            'SBL1': 'sbl1.mbn',
            # ICE16 firmware files
            'RestorePSI': 'restorepsi.bin',
            'PSI': 'psi_ram.bin',
            # ICE19 firmware files
            'RestorePSI2': 'restorepsi2.bin',
            'PSI2': 'psi_ram2.bin',
            # Mav20 Firmware file
            'Misc': 'multi_image.mbn',
        }
        return bbfw_fn_elem.get(elem)

    def fls_parse(self, buffer):
        raise NotImplementedError()

    def fls_update_sig_blob(self, buffer, blob):
        raise NotImplementedError()

    def fls_insert_ticket(self, fls, bbticket):
        raise NotImplementedError()

    def sign_bbfw(self, bbfw_orig, bbtss, bb_nonce):
        # check for BBTicket in result
        bbticket = bbtss.bb_ticket
        bbfw_dict = bbtss.get('BasebandFirmware')
        is_fls = False
        signed_file = []

        with tempfile.NamedTemporaryFile() as tmp_zip_read:
            with tempfile.NamedTemporaryFile() as tmp_zip_write:
                bbfw_patched = zipfile.ZipFile(tmp_zip_write, 'w')

                tmp_zip_read.write(bbfw_orig)
                bbfw_orig = zipfile.ZipFile(tmp_zip_read.name, 'r')

                for key, blob in bbfw_dict.items():
                    if key.endswith('-Blob') and isinstance(blob, bytes):
                        key = key.split('-', 1)[0]
                        signfn = self.get_bbfw_fn_for_element(key)

                        if signfn is None:
                            raise PyMobileDevice3Exception(
                                f'can\'t match element name \'{key}\' to baseband firmware file name.')

                        if signfn.endswith('.fls'):
                            is_fls = True

                        buffer = bbfw_orig.read(signfn)

                        if is_fls:
                            fls = self.fls_parse(buffer)
                            data = self.fls_update_sig_blob(fls, blob)
                        else:
                            parsed_sig_offset = len(buffer) - len(blob)
                            data = buffer[:parsed_sig_offset] + blob

                        bbfw_patched.writestr(bbfw_orig.getinfo(signfn), data)

                        if is_fls and (bb_nonce is None):
                            if key == 'RamPSI':
                                signed_file.append(signfn)
                        else:
                            signed_file.append(signfn)

                # remove everything but required files
                for entry in bbfw_orig.filelist:
                    keep = False
                    filename = entry.filename

                    if filename in signed_file:
                        keep = True

                    # check for anything but .mbn and .fls if bb_nonce is set
                    if bb_nonce and not keep:
                        ext = os.path.splitext(filename)[1]
                        keep |= ext in ('.fls', '.mbn', '.elf', '.bin')

                    if keep and (filename not in signed_file):
                        bbfw_patched.writestr(bbfw_orig.getinfo(filename), bbfw_orig.read(filename))

                if bb_nonce:
                    if is_fls:
                        # add BBTicket to file ebl.fls
                        buffer = bbfw_orig.read('ebl.fls')
                        fls = self.fls_parse(buffer)
                        data = self.fls_insert_ticket(fls, bbticket)
                        bbfw_patched.writestr('ebl.fls', data)
                    else:
                        # add BBTicket as bbticket.der
                        zname = zipfile.ZipInfo('bbticket.der')
                        zname.filename = 'bbticket.der'
                        ZIP_EXT_ATTR_FILE = 0o100000
                        zname.external_attr = (0o644 | ZIP_EXT_ATTR_FILE) << 16
                        bbfw_patched.writestr(zname, bbticket)

                bbfw_patched.close()
                tmp_zip_write.seek(0)
                return tmp_zip_write.read()

    def send_baseband_data(self, message: Mapping):
        self.logger.info(f'About to send BasebandData: {message}')

        # NOTE: this function is called 2 or 3 times!

        # setup request data
        arguments = message['Arguments']
        bb_chip_id = arguments.get('ChipID')
        bb_cert_id = arguments.get('CertID')
        bb_snum = arguments.get('ChipSerialNo')
        bb_nonce = arguments.get('Nonce')
        bbtss = self.bbtss

        if (bb_nonce is None) or (self.bbtss is None):
            # populate parameters
            parameters = {'ApECID': self.device.ecid}
            if bb_nonce:
                parameters['BbNonce'] = bb_nonce
            parameters['BbChipID'] = bb_chip_id
            parameters['BbGoldCertId'] = bb_cert_id
            parameters['BbSNUM'] = bb_snum

            self.build_identity.populate_tss_request_parameters(parameters)

            # create baseband request
            request = TSSRequest()

            # add baseband parameters
            request.add_common_tags(parameters)
            request.add_baseband_tags(parameters)

            fdr_support = self.build_identity['Info'].get('FDRSupport', False)
            if fdr_support:
                request.update({'ApProductionMode': True, 'ApSecurityMode': True})

            self.logger.info('Sending Baseband TSS request...')
            bbtss = request.send_receive()

            if bb_nonce:
                # keep the response for later requests
                self.bbtss = bbtss

        # get baseband firmware file path from build identity
        bbfwpath = self.build_identity['Manifest']['BasebandFirmware']['Info']['Path']

        # extract baseband firmware to temp file
        bbfw = self.ipsw.read(bbfwpath)

        buffer = self.sign_bbfw(bbfw, bbtss, bb_nonce)

        self.logger.info('Sending BasebandData now...')
        self._restored.send({'BasebandData': buffer})

    def send_fdr_trust_data(self, message):
        self.logger.info('About to send FDR Trust data...')

        # FIXME: What should we send here?
        # Sending an empty dict makes it continue with FDR
        # and this is what iTunes seems to be doing too
        self.logger.info('Sending FDR Trust data now...')
        self._restored.send({})

    def send_image_data(self, message, image_list_k, image_type_k, image_data_k):
        self.logger.debug(f'send_image_data: {message}')
        arguments = message['Arguments']
        want_image_list = arguments.get(image_list_k)
        image_name = arguments.get('ImageName')
        build_id_manifest = self.build_identity['Manifest']

        if not want_image_list and image_name is not None:
            if image_name not in build_id_manifest:
                if image_name.startswith('Ap'):
                    image_name = image_name.replace('Ap', 'Ap,')
                    if image_name not in build_id_manifest:
                        raise PyMobileDevice3Exception(f'{image_name} not in build_id_manifest')

        if image_type_k is None:
            image_type_k = arguments['ImageType']

        if image_type_k is None:
            raise PyMobileDevice3Exception('missing ImageType')

        if want_image_list is None and image_name is None:
            self.logger.info(f'About to send {image_data_k}...')

        matched_images = []
        data_dict = dict()

        for component, manifest_entry in build_id_manifest.items():
            if not isinstance(manifest_entry, dict):
                continue

            is_image_type = manifest_entry['Info'].get(image_type_k)
            if is_image_type:
                if want_image_list:
                    self.logger.info(f'found {component} component')
                    matched_images.append(component)
                elif image_name is None or image_name == component:
                    if image_name is None:
                        self.logger.info(f'found {image_type_k} component \'{component}\'')
                    else:
                        self.logger.info(f'found component \'{component}\'')

                    data_dict[component] = self.build_identity.get_component(component,
                                                                             tss=self.recovery.tss).personalized_data

        req = dict()
        if want_image_list:
            req[image_list_k] = matched_images
            self.logger.info(f'Sending {image_type_k} image list')
        else:
            if image_name:
                if image_name in data_dict:
                    req[image_data_k] = data_dict[image_name]
                req['ImageName'] = image_name
                self.logger.info(f'Sending {image_type_k} for {image_name}...')
            else:
                req[image_data_k] = data_dict
                self.logger.info(f'Sending {image_type_k} now...')

        self._restored.send(req)

    def send_bootability_bundle_data(self, message):
        self.logger.debug(f'send_bootability_bundle_data: {message}')
        data_port = message['DataPort']
        self.logger.info('Connecting to BootabilityBundle data port')

        while True:
            try:
                client = ServiceConnection.create_using_usbmux(self._restored.udid, data_port)
                break
            except ConnectionFailedError:
                self.logger.debug('Retrying connection...')

        if not client:
            raise ConnectionFailedError(f'failed to establish connection to {data_port}')

        self.logger.info('Connected to BootabilityBundle data port')

        client.sendall(self.ipsw.bootability)
        client.close()

    def send_manifest(self):
        self.logger.debug('send_manifest')
        self._restored.send({'ReceiptManifest': self.build_identity.manifest})

    def get_se_firmware_data(self, info: Mapping):
        chip_id = info.get('SE,ChipID')
        if chip_id is None:
            chip_id = info.get('SEChipID')
            if chip_id is None:
                chip_id = self.build_identity['Manifest']['SEChipID']

        if chip_id == 0x20211:
            comp_name = 'SE,Firmware'
        elif chip_id in (0x73, 0x64, 0xC8, 0xD2):
            comp_name = 'SE,UpdatePayload'
        else:
            self.logger.warning(f'Unknown SE,ChipID {chip_id} detected. Restore might fail.')

            if self.build_identity.has_component('SE,UpdatePayload'):
                comp_name = 'SE,UpdatePayload'
            elif self.build_identity.has_component('SE,Firmware'):
                comp_name = 'SE,Firmware'
            else:
                raise NotImplementedError('Neither \'SE,Firmware\' nor \'SE,UpdatePayload\' found in build identity.')

        component_data = self.build_identity.get_component(comp_name).data

        # create SE request
        request = TSSRequest()
        parameters = dict()

        # add manifest for current build_identity to parameters
        self.build_identity.populate_tss_request_parameters(parameters)

        # add SE,* tags from info dictionary to parameters
        parameters.update(info)

        # add required tags for SE TSS request
        request.add_se_tags(parameters, None)

        self.logger.info('Sending SE TSS request...')
        response = request.send_receive()

        if 'SE,Ticket' in response:
            self.logger.info('Received SE ticket')
        else:
            raise PyMobileDevice3Exception('No \'SE,Ticket\' in TSS response, this might not work')

        response['FirmwareData'] = component_data

        return response

    def get_yonkers_firmware_data(self, info: Mapping):
        # create Yonkers request
        request = TSSRequest()
        parameters = dict()

        # add manifest for current build_identity to parameters
        self.build_identity.populate_tss_request_parameters(parameters)

        # add Yonkers,* tags from info dictionary to parameters
        parameters.update(info)

        # add required tags for Yonkers TSS request
        comp_name = request.add_yonkers_tags(parameters, None)

        if comp_name is None:
            raise PyMobileDevice3Exception('Could not determine Yonkers firmware component')

        self.logger.debug(f'restore_get_yonkers_firmware_data: using {comp_name}')

        self.logger.info('Sending SE Yonkers request...')
        response = request.send_receive()

        if 'Yonkers,Ticket' in response:
            self.logger.info('Received SE ticket')
        else:
            raise PyMobileDevice3Exception('No \'Yonkers,Ticket\' in TSS response, this might not work')

        # now get actual component data
        component_data = self.build_identity.get_component(comp_name).data

        firmware_data = {
            'YonkersFirmware': component_data,
        }

        response['FirmwareData'] = firmware_data

        return response

    def get_savage_firmware_data(self, info: Mapping):
        # create Savage request
        request = TSSRequest()
        parameters = dict()

        # add manifest for current build_identity to parameters
        self.build_identity.populate_tss_request_parameters(parameters)

        # add Savage,* tags from info dictionary to parameters
        parameters.update(info)

        # add required tags for Savage TSS request
        comp_name = request.add_savage_tags(parameters, None)

        if comp_name is None:
            raise PyMobileDevice3Exception('Could not determine Savage firmware component')

        self.logger.debug(f'restore_get_savage_firmware_data: using {comp_name}')

        self.logger.info('Sending SE Savage request...')
        response = request.send_receive()

        if 'Savage,Ticket' in response:
            self.logger.info('Received SE ticket')
        else:
            raise PyMobileDevice3Exception('No \'Savage,Ticket\' in TSS response, this might not work')

        # now get actual component data
        component_data = self.build_identity.get_component(comp_name).data
        component_data = struct.pack('<L', len(component_data)) + b'\x00' * 12

        response['FirmwareData'] = component_data

        return response

    def get_rose_firmware_data(self, info: Mapping):
        self.logger.info(f'get_rose_firmware_data: {info}')

        # create Rose request
        request = TSSRequest()
        parameters = dict()

        # add manifest for current build_identity to parameters
        self.build_identity.populate_tss_request_parameters(parameters)

        parameters['ApProductionMode'] = True

        if self.device.is_image4_supported:
            parameters['ApSecurityMode'] = True
            parameters['ApSupportsImg4'] = True
        else:
            parameters['ApSupportsImg4'] = False

        # add Rap,* tags from info dictionary to parameters
        parameters.update(info)

        # add required tags for Rose TSS request
        request.add_rose_tags(parameters, None)

        self.logger.info('Sending Rose TSS request...')
        response = request.send_receive()

        rose_ticket = response.get('Rap,Ticket')
        if rose_ticket is None:
            self.logger.error('No "Rap,Ticket" in TSS response, this might not work')

        comp_name = 'Rap,RTKitOS'
        component_data = self.build_identity.get_component(comp_name).data

        ftab = Ftab(component_data)

        comp_name = 'Rap,RestoreRTKitOS'
        if self.build_identity.has_component(comp_name):
            rftab = Ftab(self.build_identity.get_component(comp_name).data)

            component_data = rftab.get_entry_data(b'rrko')
            if component_data is None:
                self.logger.error('Could not find "rrko" entry in ftab. This will probably break things')
            else:
                ftab.add_entry(b'rrko', component_data)

        response['FirmwareData'] = ftab.data

        return response

    def get_veridian_firmware_data(self, info: Mapping):
        self.logger.info(f'get_veridian_firmware_data: {info}')
        comp_name = 'BMU,FirmwareMap'

        # create Veridian request
        request = TSSRequest()
        parameters = dict()

        # add manifest for current build_identity to parameters
        self.build_identity.populate_tss_request_parameters(parameters)

        # add BMU,* tags from info dictionary to parameters
        parameters.update(info)

        # add required tags for Veridian TSS request
        request.add_veridian_tags(parameters, None)

        self.logger.info('Sending Veridian TSS request...')
        response = request.send_receive()

        ticket = response.get('BMU,Ticket')
        if ticket is None:
            self.logger.warning('No "BMU,Ticket" in TSS response, this might not work')

        component_data = self.build_identity.get_component(comp_name).data
        fw_map = plistlib.loads(component_data)
        fw_map['fw_map_digest'] = self.build_identity['Manifest'][comp_name]['Digest']

        bin_plist = plistlib.dumps(fw_map, fmt=plistlib.PlistFormat.FMT_BINARY)
        response['FirmwareData'] = bin_plist

        return response

    def get_tcon_firmware_data(self, info: Mapping):
        self.logger.info(f'restore_get_tcon_firmware_data: {info}')
        comp_name = 'Baobab,TCON'

        # create Baobab request
        request = TSSRequest()
        parameters = dict()

        # add manifest for current build_identity to parameters
        self.build_identity.populate_tss_request_parameters(parameters)

        # add Baobab,* tags from info dictionary to parameters
        parameters.update(info)

        # add required tags for Baobab TSS request
        request.add_tcon_tags(parameters, None)

        self.logger.info('Sending Baobab TSS request...')
        response = request.send_receive()

        ticket = response.get('Baobab,Ticket')
        if ticket is None:
            self.logger.warning('No "Baobab,Ticket" in TSS response, this might not work')

        response['FirmwareData'] = self.build_identity.get_component(comp_name).data

        return response

    def get_cryptex1_firmware_data(self, updater_name: str, info: Mapping, arguments: Mapping):
        self.logger.info(f'get_cryptex1_firmware_data: {arguments}')
        request = TSSRequest()
        parameters = dict()

        # add manifest for current build_identity to parameters
        self.build_identity.populate_tss_request_parameters(
            parameters, arguments['DeviceGeneratedTags']['BuildIdentityTags'])

        parameters['@BBTicket'] = True
        parameters['ApProductionMode'] = arguments['MessageArgInfo']['ApProductionMode']
        parameters['ApSecurityMode'] = True
        response_ticket = arguments['DeviceGeneratedTags']['ResponseTags'][0]

        parameters.update(arguments['DeviceGeneratedRequest'])
        request.add_common_tags(info)
        request.update(parameters)

        for redacted_field in ('RequiresUIDMode', ):
            request.remove_key(redacted_field)

        self.logger.info(f'Sending {updater_name} TSS request...')
        response = request.send_receive()

        ticket = response.get(response_ticket)
        if ticket is None:
            self.logger.warning(f'No "{response_ticket}" in TSS response, this might not work')
            self.logger.debug(response)

        return response

    def get_timer_firmware_data(self, info: Mapping):
        self.logger.info(f'get_timer_firmware_data: {info}')

        ftab = None

        # create Timer request
        request = TSSRequest()
        parameters = dict()

        # add manifest for current build_identity to parameters
        self.build_identity.populate_tss_request_parameters(parameters)

        parameters['ApProductionMode'] = True
        if self.device.is_image4_supported:
            parameters['ApSecurityMode'] = True
            parameters['ApSupportsImg4'] = True
        else:
            parameters['ApSupportsImg4'] = False

        # add Timer,* tags from info dictionary to parameters
        info_array = info['InfoArray']
        info_dict = info_array[0]
        hwid = info_dict['HardwareID']
        tag = info_dict['TagNumber']
        parameters['TagNumber'] = tag
        ticket_name = info_dict['TicketName']
        parameters['TicketName'] = ticket_name
        parameters[f'Timer,ChipID,{tag}'] = hwid['ChipID']
        parameters[f'Timer,BoardID,{tag}'] = hwid['BoardID']
        parameters[f'Timer,ECID,{tag}'] = hwid['ECID']
        parameters[f'Timer,Nonce,{tag}'] = hwid['Nonce']
        parameters[f'Timer,SecurityMode,{tag}'] = hwid['SecurityMode']
        parameters[f'Timer,SecurityDomain,{tag}'] = hwid['SecurityDomain']
        parameters[f'Timer,ProductionMode,{tag}'] = hwid['ProductionMode']

        ap_info = info['APInfo']
        parameters.update(ap_info)

        # add required tags for Timer TSS request
        request.add_timer_tags(parameters, None)

        self.logger.info(f'Sending {ticket_name} TSS request...')
        response = request.send_receive()

        ticket = response.get(ticket_name)
        if ticket is None:
            self.logger.warning(f'No "{ticket_name}" in TSS response, this might not work')

        comp_name = f'Timer,RTKitOS,{tag}'
        if self.build_identity.has_component(comp_name):
            ftab = Ftab(self.build_identity.get_component(comp_name).data)
            if ftab.tag != b'rkos':
                self.logger.warning(f'Unexpected tag {ftab.tag}. continuing anyway.')
        else:
            self.logger.info(f'NOTE: Build identity does not have a "{comp_name}" component.')

        comp_name = f'Timer,RestoreRTKitOS,{tag}'
        if self.build_identity.has_component(comp_name):
            rftab = Ftab(self.build_identity.get_component(comp_name).data)

            component_data = rftab.get_entry_data(b'rrko')
            if component_data is None:
                self.logger.error('Could not find "rrko" entry in ftab. This will probably break things')
            else:
                if ftab is None:
                    raise PyMobileDevice3Exception('ftab is None')
                ftab.add_entry(b'rrko', component_data)
        else:
            self.logger.info(f'NOTE: Build identity does not have a "{comp_name}" component.')

        response['FirmwareData'] = ftab.data

        return response

    def send_firmware_updater_data(self, message: Mapping):
        self.logger.debug(f'got FirmwareUpdaterData request: {message}')
        arguments = message['Arguments']
        s_type = arguments['MessageArgType']
        updater_name = arguments['MessageArgUpdaterName']

        if s_type not in ('FirmwareResponseData',):
            raise PyMobileDevice3Exception(f'MessageArgType has unexpected value \'{s_type}\'')

        info = arguments['MessageArgInfo']

        if updater_name == 'SE':
            fwdict = self.get_se_firmware_data(info)
            if fwdict is None:
                raise PyMobileDevice3Exception('Couldn\'t get SE firmware data')

        elif updater_name == 'Savage':
            fwtype = 'Savage'
            info2 = info.get('YonkersDeviceInfo')
            if info2:
                fwtype = 'Yonkers'
                fwdict = self.get_yonkers_firmware_data(info2)
            else:
                fwdict = self.get_savage_firmware_data(info)
            if fwdict is None:
                raise PyMobileDevice3Exception(f'Couldn\'t get {fwtype} firmware data')

        elif updater_name == 'Rose':
            fwdict = self.get_rose_firmware_data(info)
            if fwdict is None:
                raise PyMobileDevice3Exception('Couldn\'t get Rose firmware data')

        elif updater_name == 'T200':
            fwdict = self.get_veridian_firmware_data(info)
            if fwdict is None:
                raise PyMobileDevice3Exception('Couldn\'t get Veridian firmware data')

        elif updater_name == 'AppleTCON':
            fwdict = self.get_tcon_firmware_data(info)
            if fwdict is None:
                raise PyMobileDevice3Exception('Couldn\'t get TCON firmware data')

        elif updater_name == 'AppleTypeCRetimer':
            fwdict = self.get_timer_firmware_data(info)
            if fwdict is None:
                raise PyMobileDevice3Exception('Couldn\'t get AppleTypeCRetimer firmware data')

        elif updater_name in ('Cryptex1', 'Cryptex1LocalPolicy'):
            fwdict = self.get_cryptex1_firmware_data(updater_name, info, arguments)
            if fwdict is None:
                raise PyMobileDevice3Exception(f'Couldn\'t get {updater_name} firmware data')

        else:
            raise PyMobileDevice3Exception(f'Got unknown updater name: {updater_name}')

        self.logger.info('Sending FirmwareResponse data now...')
        self._restored.send({'FirmwareResponseData': fwdict})

    def send_firmware_updater_preflight(self, message: Mapping):
        self.logger.warning(f'send_firmware_updater_preflight: {message}')
        self._restored.send({})

    def send_component(self, component, component_name=None):
        if component_name is None:
            component_name = component

        self.logger.info(f'Sending now {component_name}...')
        self._restored.send(
            {f'{component_name}File': self.build_identity.get_component(component,
                                                                        tss=self.recovery.tss).personalized_data})

    def handle_data_request_msg(self, message: Mapping):
        self.logger.info(f'handle_data_request_msg: {message}')

        # checks and see what kind of data restored is requests and pass the request to its own handler
        data_type = message.get('DataType')

        if not isinstance(data_type, str):
            return

        if data_type in self._data_request_handlers:
            self._data_request_handlers[data_type](message)
        elif data_type in self._data_request_components:
            self._data_request_components[data_type](data_type)
        elif data_type == 'SystemImageRootHash':
            self.send_component('SystemVolume', data_type)
        elif data_type == 'SystemImageCanonicalMetadata':
            self.send_component('Ap,SystemVolumeCanonicalMetadata', data_type)
        elif data_type == 'FUDData':
            self.send_image_data(message, 'FUDImageList', 'IsFUDFirmware', 'FUDImageData')
        elif data_type == 'PersonalizedData':
            self.send_image_data(message, 'ImageList', None, 'ImageData')
        elif data_type == 'EANData':
            self.send_image_data(message, 'EANImageList', 'IsEarlyAccessFirmware', 'EANData')
        elif data_type == 'BootabilityBundle':
            self.send_bootability_bundle_data(message)
        elif data_type == 'ReceiptManifest':
            self.send_manifest()
        elif data_type == 'BasebandUpdaterOutputData':
            self.handle_baseband_updater_output_data(message)
        else:
            self.logger.error(f'unknown data request: {message}')

    def handle_previous_restore_log_msg(self, message: Mapping):
        restorelog = message['PreviousRestoreLog']
        self.logger.info(f'PreviousRestoreLog: {restorelog}')

    def handle_progress_msg(self, message: Mapping):
        operation = message['Operation']
        if operation in PROGRESS_BAR_OPERATIONS:
            message['Operation'] = PROGRESS_BAR_OPERATIONS[operation]

        if message['Operation'] == 'VERIFY_RESTORE':
            progress = message['Progress']

            if self._pb_verify_restore is None:
                self._pb_verify_restore = tqdm(total=100, desc='verify-restore', dynamic_ncols=True)
                self._pb_verify_restore_old_value = 0

            self._pb_verify_restore.update(progress - self._pb_verify_restore_old_value)
            self._pb_verify_restore_old_value = progress

            if progress == 100:
                self._pb_verify_restore.close()
                self._pb_verify_restore = None

            return

        self.logger.info(f'progress-bar: {message}')

    def handle_status_msg(self, message: Mapping):
        self.logger.debug(f'status message: {message}')
        status = message['Status']
        log = message.get('Log')

        if log:
            # this is the true device log that may inform us for anything that went wrong
            # we want it to be output into the stdout in multiline, so we can inspect it using
            # easier shell commands
            print(log)

        if status == 0:
            self._restore_finished = True
            self._restored.send({'MsgType': 'ReceivedFinalStatusMsg'})
        else:
            if status in known_errors:
                self.logger.error(known_errors[status])
            else:
                self.logger.error('unknown error')

    def handle_checkpoint_msg(self, message: Mapping):
        self.logger.info(f'checkpoint: {message}')

    def handle_bb_update_status_msg(self, message: Mapping):
        self.logger.info(f'bb_update_status_msg: {message}')
        if not message['Accepted']:
            raise PyMobileDevice3Exception(str(message))

    def handle_baseband_updater_output_data(self, message: Mapping):
        self.logger.debug(f'restore_handle_baseband_updater_output_data: {message}')
        data_port = message['DataPort']

        self.logger.info('Connecting to baseband updater data port')

        while True:
            try:
                client = ServiceConnection.create_using_usbmux(self._restored.udid, data_port)
                break
            except ConnectionFailedError:
                self.logger.debug('Retrying connection...')

        if not client:
            raise ConnectionFailedError(f'failed to establish connection to {data_port}')

        self.logger.info('Connected to BasebandUpdaterOutputData data port')

        filename = f'updater_output-{self._restored.udid}.cpio'
        self.logger.info(f'Writing updater output into: {filename}')

        with open(filename, 'wb') as f:
            while True:
                buf = client.recv()
                if not buf:
                    break
                f.write(buf)

        self.logger.debug('Closing connection of BasebandUpdaterOutputData data port')
        client.close()

    def _connect_to_restored_service(self):
        while True:
            try:
                self._restored = RestoredClient()
                break
            except NoDeviceConnectedError:
                pass

    def restore_device(self):
        self.logger.debug('waiting for device to connect for restored service')
        self._connect_to_restored_service()

        self.logger.info(f'hardware info: {self._restored.hardware_info}')
        self.logger.info(f'version: {self._restored.version}')
        self.logger.info(f'saved_debug_info: {self._restored.saved_debug_info}')

        if self.recovery.tss.bb_ticket is not None:
            # initial TSS response contains a baseband ticket
            self.bbtss = self.recovery.tss

        if self._ignore_fdr:
            self.logger.info('Establishing a mock FDR listener')
            self._fdr = ServiceConnection.create_using_usbmux(self._restored.udid, FDRClient.SERVICE_PORT)
        else:
            self.logger.info('Starting FDR listener thread')
            start_fdr_thread(fdr_type.FDR_CTRL)

        sep = self.build_identity['Manifest']['SEP'].get('Info')
        spp = self.build_identity['Info'].get('SystemPartitionPadding')
        opts = RestoreOptions(preflight_info=self._preflight_info, sep=sep, macos_variant=self.macos_variant,
                              build_identity=self.build_identity, restore_boot_args=self.recovery.restore_boot_args,
                              spp=spp, restore_behavior=self.build_identity.restore_behavior,
                              msp=self.build_identity.minimum_system_partition)

        # start the restore process
        self._restored.start_restore(opts)

        # this is the restore process loop, it reads each message in from
        # restored and passes that data on to its specific handler
        while not self._restore_finished:
            # finally, if any of these message handlers returned -1 then we encountered
            # an unrecoverable error, so we need to bail.
            message = self._restored.recv()

            # discover what kind of message has been received
            message_type = message.get('MsgType')

            if message_type in self._handlers:
                try:
                    self._handlers[message_type](message)
                except Exception:
                    traceback.print_exc()
            else:
                # there might be some other message types i'm not aware of, but I think
                # at least the "previous error logs" messages usually end up here
                self.logger.debug(f'unhandled message type received: {message}')

    def update(self):
        self.recovery.boot_ramdisk()

        # device is finally in restore mode, let's do this
        self.restore_device()



# extracted from ac2
import logging
import uuid

from pymobiledevice3.restore.ipsw.build_identity import BuildIdentity

logger = logging.getLogger(__name__)

SUPPORTED_DATA_TYPES = {
    'BasebandBootData': False,
    'BasebandData': False,
    'BasebandStackData': False,
    'BasebandUpdaterOutputData': False,
    'BootabilityBundle': False,
    'BuildIdentityDict': False,
    'BuildIdentityDictV2': False,
    'DataType': False,
    'DiagData': False,
    'EANData': False,
    'FDRMemoryCommit': False,
    'FDRTrustData': False,
    'FUDData': False,
    'FileData': False,
    'FileDataDone': False,
    'FirmwareUpdaterData': False,
    'GrapeFWData': False,
    'HPMFWData': False,
    'HostSystemTime': True,
    'KernelCache': False,
    'NORData': False,
    'NitrogenFWData': True,
    'OpalFWData': False,
    'OverlayRootDataCount': False,
    'OverlayRootDataForKey': True,
    'PeppyFWData': True,
    'PersonalizedBootObjectV3': False,
    'PersonalizedData': True,
    'ProvisioningData': False,
    'RamdiskFWData': True,
    'RecoveryOSASRImage': True,
    'RecoveryOSAppleLogo': True,
    'RecoveryOSDeviceTree': True,
    'RecoveryOSFileAssetImage': True,
    'RecoveryOSIBEC': True,
    'RecoveryOSIBootFWFilesImages': True,
    'RecoveryOSImage': True,
    'RecoveryOSKernelCache': True,
    'RecoveryOSLocalPolicy': True,
    'RecoveryOSOverlayRootDataCount': False,
    'RecoveryOSRootTicketData': True,
    'RecoveryOSStaticTrustCache': True,
    'RecoveryOSVersionData': True,
    'RootData': False,
    'RootTicket': False,
    'S3EOverride': False,
    'SourceBootObjectV3': False,
    'SourceBootObjectV4': False,
    'SsoServiceTicket': False,
    'StockholmPostflight': False,
    'SystemImageCanonicalMetadata': False,
    'SystemImageData': False,
    'SystemImageRootHash': False,
    'USBCFWData': False,
    'USBCOverride': False,
    'FirmwareUpdaterPreflight': True,
    'ReceiptManifest': True,
    'FirmwareUpdaterDataV2': False,
    'RestoreLocalPolicy': True,
    'AuthInstallCACert': True,
    'OverlayRootDataForKeyIndex': True,
}

# extracted from ac2
SUPPORTED_MESSAGE_TYPES = {
    'BBUpdateStatusMsg': False,
    'CheckpointMsg': True,
    'DataRequestMsg': False,
    'FDRSubmit': True,
    'MsgType': False,
    'PreviousRestoreLogMsg': False,
    'ProgressMsg': False,
    'ProvisioningAck': False,
    'ProvisioningInfo': False,
    'ProvisioningStatusMsg': False,
    'ReceivedFinalStatusMsg': False,
    'RestoredCrash': True,
    'StatusMsg': False,
}


class RestoreOptions:

    def __init__(self, preflight_info=None, sep=None, macos_variant=None, build_identity: BuildIdentity = None,
                 restore_boot_args=None, spp=None, restore_behavior: str = None, msp=None):
        self.AutoBootDelay = 0

        if preflight_info is not None:
            bbus = dict(preflight_info)
            bbus.pop('FusingStatus')
            bbus.pop('PkHash')
            self.BBUpdaterState = bbus

            nonce = preflight_info.get('Nonce')
            if nonce is not None:
                self.BasebandNonce = nonce

        self.SupportedDataTypes = SUPPORTED_DATA_TYPES
        self.SupportedMessageTypes = SUPPORTED_MESSAGE_TYPES

        # FIXME: Should be adjusted for update behaviors
        if macos_variant:
            self.AddSystemPartitionPadding = True
            self.AllowUntetheredRestore = False
            self.AuthInstallEnableSso = False

            macos_variant = build_identity.macos_variant
            if macos_variant is not None:
                self.AuthInstallRecoveryOSVariant = macos_variant

            self.AuthInstallRestoreBehavior = restore_behavior
            self.AutoBootDelay = 0
            self.BasebandUpdaterOutputPath = True
            self.DisableUserAuthentication = True
            self.FitSystemPartitionToContent = True
            self.FlashNOR = True
            self.FormatForAPFS = True
            self.FormatForLwVM = False
            self.InstallDiags = False
            self.InstallRecoveryOS = True
            self.MacOSSwapPerformed = True
            self.MacOSVariantPresent = True
            self.MinimumBatteryVoltage = 0  # FIXME: Should be adjusted for M1 macbooks (if needed)
            self.RecoveryOSUnpack = True
            self.ShouldRestoreSystemImage = True
            self.SkipPreflightPersonalization = False
            self.UpdateBaseband = True

            # FIXME: I don't know where this number comes from yet.
            #  It seems like it matches this part of the build identity:
            # 	<key>OSVarContentSize</key>
            # 	<integer>573751296</integer>
            # It did work with multiple macOS versions
            self.recoveryOSPartitionSize = 58201
            if msp:
                self.SystemPartitionSize = msp
        else:
            self.BootImageType = 'UserOrInternal'
            self.DFUFileType = 'RELEASE'
            self.DataImage = False
            self.FirmwareDirectory = '.'
            self.FlashNOR = True
            self.KernelCacheType = 'Release'
            self.NORImageType = 'production'
            self.RestoreBundlePath = '/tmp/Per2.tmp'
            self.SystemImageType = 'User'
            self.UpdateBaseband = False

            if sep is not None:
                required_capacity = sep.get('RequiredCapacity')
                if required_capacity:
                    logger.debug(f'TZ0RequiredCapacity: {required_capacity}')
                    self.TZ0RequiredCapacity = required_capacity

            self.PersonalizedDuringPreflight = True

        self.RootToInstall = False
        self.UUID = str(uuid.uuid4()).upper()
        self.CreateFilesystemPartitions = True
        self.SystemImage = True

        if restore_boot_args is not None:
            self.RestoreBootArgs = restore_boot_args

        if spp:
            spp = dict(spp)
        else:
            spp = {'128': 1280, '16': 160, '32': 320, '64': 640, '8': 80}
        self.SystemPartitionPadding = spp

    def to_dict(self):
        return self.__dict__



import hashlib
import logging
import time
import typing
from io import BytesIO

from usb import USBError

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.irecv import IRecv, Mode
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.restore.base_restore import BaseRestore, Behavior
from pymobiledevice3.restore.consts import lpol_file
from pymobiledevice3.restore.device import Device
from pymobiledevice3.restore.tss import TSSRequest

RESTORE_VARIANT_ERASE_INSTALL = 'Erase Install (IPSW)'
RESTORE_VARIANT_UPGRADE_INSTALL = 'Upgrade Install (IPSW)'
RESTORE_VARIANT_MACOS_RECOVERY_OS = 'macOS Customer'


class Recovery(BaseRestore):
    def __init__(self, ipsw: BytesIO, device: Device, tss: typing.Mapping = None, behavior: Behavior = Behavior.Update):
        super().__init__(ipsw, device, tss, behavior, logger=logging.getLogger(__name__))
        self.tss_localpolicy = None
        self.tss_recoveryos_root_ticket = None
        self.restore_boot_args = None

    def reconnect_irecv(self, is_recovery=None):
        self.logger.debug('waiting for device to reconnect...')
        self.device.irecv = IRecv(ecid=self.device.ecid, is_recovery=is_recovery)
        self.logger.debug(f'connected mode: {self.device.irecv.mode}')

    def get_preboard_manifest(self):
        overrides = {
            '@APTicket': True,
            'ApProductionMode': 0,
            'ApSecurityDomain': 0,
        }

        parameters = {
            'ApProductionMode': False,
            'ApSecurityMode': False,
            'ApSupportsImg4': True,
        }

        self.build_identity.populate_tss_request_parameters(parameters)

        tss = TSSRequest()
        tss.add_common_tags(parameters, overrides)

        parameters['_OnlyFWComponents'] = True

        tss.add_ap_tags(parameters)

        return tss.img4_create_local_manifest(build_identity=self.build_identity)

    def get_tss_response(self):
        # populate parameters
        parameters = dict()

        parameters['ApECID'] = self.device.ecid
        if self.device.ap_nonce is not None:
            parameters['ApNonce'] = self.device.ap_nonce

        if self.device.sep_nonce is not None:
            parameters['ApSepNonce'] = self.device.sep_nonce

        parameters['ApProductionMode'] = True

        if self.device.is_image4_supported:
            parameters['ApSecurityMode'] = True
            parameters['ApSupportsImg4'] = True
        else:
            parameters['ApSupportsImg4'] = False

        self.build_identity.populate_tss_request_parameters(parameters)

        tss = TSSRequest()
        tss.add_common_tags(parameters)
        tss.add_ap_tags(parameters)

        # add personalized parameters
        if self.device.is_image4_supported:
            tss.add_ap_img4_tags(parameters)
        else:
            tss.add_ap_img3_tags(parameters)

        # normal mode; request baseband ticket as well
        if self.device.lockdown is not None:
            pinfo = self.device.lockdown.preflight_info
            if pinfo:
                self.logger.debug('adding preflight info')

                node = pinfo.get('Nonce')
                if node is not None:
                    parameters['BbNonce'] = node

                node = pinfo.get('ChipID')
                if node is not None:
                    parameters['BbChipID'] = node

                node = pinfo.get('CertID')
                if node is not None:
                    parameters['BbGoldCertId'] = node

                node = pinfo.get('ChipSerialNo')
                if node is not None:
                    parameters['BbSNUM'] = node

                tss.add_baseband_tags(parameters)

                euiccchipid = pinfo.get('EUICCChipID')
                if euiccchipid:
                    self.logger.debug('adding EUICCChipID info')
                    parameters['eUICC,ChipID'] = euiccchipid

                    if euiccchipid >= 5:
                        node = pinfo.get('EUICCCSN')
                        if node is not None:
                            parameters['eUICC,EID'] = node

                        node = pinfo.get('EUICCCertIdentifier')
                        if node is not None:
                            parameters['eUICC,RootKeyIdentifier'] = node

                        node = pinfo.get('EUICCGoldNonce')
                        if node is not None:
                            parameters['EUICCGoldNonce'] = node

                        node = pinfo.get('EUICCMainNonce')
                        if node is not None:
                            parameters['EUICCMainNonce'] = node

                        tss.add_vinyl_tags(parameters)

        # send request and grab response
        return tss.send_receive()

    def get_local_policy_tss_response(self):
        # populate parameters
        parameters = {
            'ApECID': self.device.ecid,
            'Ap,LocalBoot': False,
            'ApProductionMode': True,
        }

        if self.device.ap_nonce:
            parameters['ApNonce'] = self.device.ap_nonce

        sep_nonce = self.device.sep_nonce

        if sep_nonce:
            parameters['ApSepNonce'] = sep_nonce

        if self.device.is_image4_supported:
            parameters['ApSecurityMode'] = True
            parameters['ApSupportsImg4'] = True
        else:
            parameters['ApSupportsImg4'] = False

        self.build_identity.populate_tss_request_parameters(parameters)

        # Add Ap,LocalPolicy
        lpol = {
            'Digest': hashlib.sha384(lpol_file).digest(),
            'Trusted': True,
        }

        parameters['Ap,LocalPolicy'] = lpol

        # Add Ap,NextStageIM4MHash
        # Get previous TSS ticket
        ticket = self.tss.ap_img4_ticket
        # Hash it and add it as Ap,NextStageIM4MHash
        parameters['Ap,NextStageIM4MHash'] = hashlib.sha384(ticket).digest()

        # create basic request
        request = TSSRequest()

        # add common tags from manifest
        request.add_local_policy_tags(parameters)

        return request.send_receive()

    def get_recoveryos_root_ticket_tss_response(self):
        # populate parameters
        parameters = {
            'ApECID': self.device.ecid,
            'Ap,LocalBoot': False,
            'ApProductionMode': True,
        }

        if self.device.ap_nonce:
            parameters['ApNonce'] = self.device.ap_nonce

        sep_nonce = self.device.sep_nonce

        if sep_nonce:
            parameters['ApSepNonce'] = sep_nonce

        if self.device.is_image4_supported:
            parameters['ApSecurityMode'] = True
            parameters['ApSupportsImg4'] = True
        else:
            parameters['ApSupportsImg4'] = False

        self.build_identity.populate_tss_request_parameters(parameters)

        # create basic request
        # Adds @HostPlatformInfo, @VersionInfo, @UUID
        request = TSSRequest()

        # add common tags from manifest
        # Adds Ap,OSLongVersion, AppNonce, @ApImg4Ticket
        request.add_ap_img4_tags(parameters)

        # add AP tags from manifest
        request.add_common_tags(parameters)

        # add AP tags from manifest
        # Fills digests & co
        request.add_ap_recovery_tags(parameters)

        return request.send_receive()

    def fetch_tss_record(self):
        if self.ipsw.build_manifest.build_major > 8:
            if self.device.ap_nonce is None:
                # the first nonce request with older firmware releases can fail, and it's OK
                self.logger.info('NOTE: Unable to get nonce from device')

        self.tss = self.get_tss_response()

        if self.macos_variant:
            self.tss_localpolicy = self.get_local_policy_tss_response()
            self.tss_recoveryos_root_ticket = self.get_recoveryos_root_ticket_tss_response()

        return self.tss

    def send_component(self, name: str):
        # Use a specific TSS ticket for the Ap,LocalPolicy component
        data = None
        tss = self.tss
        if name == 'Ap,LocalPolicy':
            tss = self.tss_localpolicy
            # If Ap,LocalPolicy => Inject an empty policy
            data = lpol_file

        data = self.build_identity.get_component(name, tss=tss, data=data).personalized_data
        self.logger.info(f'Sending {name} ({len(data)} bytes)...')
        self.device.irecv.send_buffer(data)

    def send_component_and_command(self, name, command):
        self.send_component(name)
        self.device.irecv.send_command(command)

    def send_ibec(self):
        component = 'iBEC'
        self.send_component(component)
        self.device.irecv.send_command('go', b_request=1)
        self.device.irecv.ctrl_transfer(0x21, 1)

    def send_applelogo(self, allow_missing=True):
        component = 'RestoreLogo'

        if not self.build_identity.has_component(component):
            if allow_missing:
                logging.warning(f'build_identity has no {component}')
                return
            else:
                raise PyMobileDevice3Exception(f'missing component: {component}')

        self.send_component(component)
        self.device.irecv.send_command('setpicture 4')
        self.device.irecv.send_command('bgcolor 0 0 0')

    def send_loaded_by_iboot(self):
        manifest = self.build_identity['Manifest']
        for key, node in manifest.items():
            iboot = node['Info'].get('IsLoadedByiBoot', False)
            iboot_stg1 = node['Info'].get('IsLoadedByiBootStage1', False)

            assert isinstance(iboot, bool)
            assert isinstance(iboot_stg1, bool)

            if iboot and not iboot_stg1:
                self.logger.debug(f'{key} is loaded by iBoot')
                self.send_component_and_command(key, 'firmware')

    def send_iboot_stage1_components(self):
        manifest = self.build_identity['Manifest']
        for key, node in manifest.items():
            iboot = node['Info'].get('IsLoadedByiBoot', False)
            iboot_stg1 = node['Info'].get('IsLoadedByiBootStage1', False)

            assert isinstance(iboot, bool)
            assert isinstance(iboot_stg1, bool)

            if iboot and iboot_stg1:
                self.logger.debug(f'{key} is loaded by iBoot Stage 1')
                self.send_component_and_command(key, 'firmware')

    def send_ramdisk(self):
        component = 'RestoreRamDisk'
        ramdisk_size = self.device.irecv.getenv('ramdisk-size')
        self.logger.info(f'ramdisk-size: {ramdisk_size}')

        self.send_component(component)
        ramdisk_delay = self.device.irecv.getenv('ramdisk-delay')
        self.logger.info(f'ramdisk-delay: {ramdisk_delay}')

        self.device.irecv.send_command('ramdisk')

        time.sleep(2)

    def send_kernelcache(self):
        component = 'RestoreKernelCache'

        self.send_component(component)
        try:
            self.device.irecv.ctrl_transfer(0x21, 1)
        except USBError:
            pass

        if self.restore_boot_args:
            self.device.irecv.send_command(f'setenv boot-args {self.restore_boot_args}')

        try:
            self.device.irecv.send_command('bootx', b_request=1)
        except USBError:
            pass

    def set_autoboot(self, enable: bool):
        self.device.irecv.set_autoboot(enable)

    def enter_restore(self):
        if self.ipsw.build_manifest.build_major >= 8:
            self.restore_boot_args = 'rd=md0 nand-enable-reformat=1 -progress'
        elif self.macos_variant:
            self.restore_boot_args = 'rd=md0 nand-enable-reformat=1 -progress -restore'

        # upload data to make device boot restore mode

        # Recovery Mode Environment:
        build_version = None
        while not build_version:
            self.logger.debug('build-version not yet supported. reconnecting...')
            time.sleep(1)

            # sometimes we manage to connect before iBEC actually started running
            build_version = self.device.irecv.getenv('build-version')
            self.reconnect_irecv()

        self.logger.info(f'iBoot build-version={build_version}')

        build_style = self.device.irecv.getenv('build-style')
        self.logger.info(f'iBoot build-style={build_style}')

        radio_error = self.device.irecv.getenv('radio-error')
        if radio_error:
            radio_error = int(radio_error)
            self.logger.info(f'radio-error: {radio_error}')
            radio_error_string = self.device.irecv.getenv('radio-error-string')
            if radio_error_string:
                self.logger.info(f'radio-error-string: {radio_error_string}')

        self.set_autoboot(False)

        # send logo and show it
        self.send_applelogo()

        # send components loaded by iBoot
        self.send_loaded_by_iboot()

        # send ramdisk and run it
        self.send_ramdisk()

        # send devicetree and load it
        self.send_component_and_command('RestoreDeviceTree', 'devicetree')

        if self.build_identity.has_component('RestoreSEP'):
            # send rsepfirmware and load it
            self.send_component_and_command('RestoreSEP', 'rsepfirmware')

        self.send_kernelcache()

    def dfu_enter_recovery(self):
        self.send_component('iBSS')
        self.reconnect_irecv(is_recovery=True)

        if self.build_identity.build_manifest.build_major > 8:
            old_nonce = self.device.irecv.ap_nonce

            # reconnect
            self.reconnect_irecv()
            nonce = self.device.irecv.ap_nonce

            if old_nonce != nonce:
                # Welcome iOS5. We have to re-request the TSS with our nonce.
                self.tss = self.get_tss_response()

            self.device.irecv.set_configuration(1)

            # Now, before sending iBEC, we must send necessary firmwares on new versions.
            if self.macos_variant:
                # Without this empty policy file & its special signature, iBEC won't start.
                self.send_component_and_command('Ap,LocalPolicy', 'lpolrestore')
                self.send_iboot_stage1_components()
                self.device.irecv.set_autoboot(False)
                self.device.irecv.send_command('setenvnp boot-args rd=md0 nand-enable-reformat=1 -progress -restore')
                self.send_applelogo(allow_missing=False)

            # send iBEC
            self.send_component('iBEC')

            if self.device.irecv and self.device.irecv.mode.is_recovery:
                time.sleep(1)
                self.device.irecv.send_command('go', b_request=1)

                if self.build_identity.build_manifest.build_major < 20:
                    self.device.irecv.ctrl_transfer(0x21, 1, timeout=5000)

        self.logger.debug('Waiting for device to disconnect...')
        time.sleep(10)

        self.logger.debug('Waiting for device to reconnect in recovery mode...')
        self.reconnect_irecv(is_recovery=True)

    def boot_ramdisk(self):
        if self.tss is None:
            self.logger.info('fetching TSS record')
            self.fetch_tss_record()

        if self.device.irecv:
            if self.device.irecv.mode == Mode.DFU_MODE:
                # device is currently in DFU mode, place it into recovery mode
                self.dfu_enter_recovery()
            elif self.device.irecv.mode.is_recovery:
                # now we load the iBEC
                try:
                    self.send_ibec()
                except USBError:
                    pass

                self.reconnect_irecv()
        elif self.device.lockdown:
            # normal mode
            self.logger.info('going into Recovery')

            # in case lockdown has disconnected while waiting for a ticket
            self.device.lockdown = LockdownClient(serial=self.device.lockdown.udid)
            self.device.lockdown.enter_recovery()

            self.device.lockdown = None
            self.device.irecv = IRecv(self.device.ecid)
            self.reconnect_irecv()

            # now we load the iBEC
            try:
                self.send_ibec()
            except USBError:
                pass

            self.reconnect_irecv()

        self.logger.info('device booted into recovery')

        # now finally do the magic to put the device into restore mode
        self.enter_restore()



import logging

import click
from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.amfi import AmfiService

logger = logging.getLogger(__name__)


@click.group()
def cli():
    """ amfi cli """
    pass


@cli.group()
def amfi():
    """ amfi options """
    pass


@amfi.command(cls=Command)
def enable_developer_mode(lockdown: LockdownClient):
    """ enable developer mode """
    AmfiService(lockdown).enable_developer_mode()


@amfi.command(cls=Command)
@click.option('--color/--no-color', default=True)
def developer_mode_status(lockdown: LockdownClient, color):
    """ query developer mode status """
    print_json(lockdown.developer_mode_status, colored=color)



#!/usr/bin/env python3

import enum
import socket
import struct

from construct import Struct, Int32ub, Int32ul, Bytes, Byte, this, Padding, Padded, CString

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.base_service import BaseService

INTERFACE_NAMES = enum.Enum('InterfaceNames', names={
    'other': 1,
    'regular1822': 2,
    'hdh1822': 3,
    'ddnX25': 4,
    'rfc877x25': 5,
    'ethernetCsmacd': 6,
    'iso88023Csmacd': 7,
    'iso88024TokenBus': 8,
    'iso88025TokenRing': 9,
    'iso88026Man': 10,
    'starLan': 11,
    'proteon10Mbit': 12,
    'proteon80Mbit': 13,
    'hyperchannel': 14,
    'fddi': 15,
    'lapb': 16,
    'sdlc': 17,
    'ds1': 18,
    'e1': 19,
    'basicISDN': 20,
    'primaryISDN': 21,
    'propPointToPointSerial': 22,
    'ppp': 23,
    'softwareLoopback': 24,
    'eon': 25,
    'ethernet3Mbit': 26,
    'nsip': 27,
    'slip': 28,
    'ultra': 29,
    'ds3': 30,
    'sip': 31,
    'frameRelay': 32,
    'rs232': 33,
    'para': 34,
    'arcnet': 35,
    'arcnetPlus': 36,
    'atm': 37,
    'miox25': 38,
    'sonet': 39,
    'x25ple': 40,
    'iso88022llc': 41,
    'localTalk': 42,
    'smdsDxi': 43,
    'frameRelayService': 44,
    'v35': 45,
    'hssi': 46,
    'hippi': 47,
    'modem': 48,
    'aal5': 49,
    'sonetPath': 50,
    'sonetVT': 51,
    'smdsIcip': 52,
    'propVirtual': 53,
    'propMultiplexor': 54,
    'ieee80212': 55,
    'fibreChannel': 56,
    'hippiInterface': 57,
    'frameRelayInterconnect': 58,
    'aflane8023': 59,
    'aflane8025': 60,
    'cctEmul': 61,
    'fastEther': 62,
    'isdn': 63,
    'v11': 64,
    'v36': 65,
    'g703at64k': 66,
    'g703at2mb': 67,
    'qllc': 68,
    'fastEtherFX': 69,
    'channel': 70,
    'ieee80211': 71,
    'ibm370parChan': 72,
    'escon': 73,
    'dlsw': 74,
    'isdns': 75,
    'isdnu': 76,
    'lapd': 77,
    'ipSwitch': 78,
    'rsrb': 79,
    'atmLogical': 80,
    'ds0': 81,
    'ds0Bundle': 82,
    'bsc': 83,
    'async': 84,
    'cnr': 85,
    'iso88025Dtr': 86,
    'eplrs': 87,
    'arap': 88,
    'propCnls': 89,
    'hostPad': 90,
    'termPad': 91,
    'frameRelayMPI': 92,
    'x213': 93,
    'adsl': 94,
    'radsl': 95,
    'sdsl': 96,
    'vdsl': 97,
    'iso88025CRFPInt': 98,
    'myrinet': 99,
    'voiceEM': 100,
    'voiceFXO': 101,
    'voiceFXS': 102,
    'voiceEncap': 103,
    'voiceOverIp': 104,
    'atmDxi': 105,
    'atmFuni': 106,
    'atmIma': 107,
    'pppMultilinkBundle': 108,
    'ipOverCdlc': 109,
    'ipOverClaw': 110,
    'stackToStack': 111,
    'virtualIpAddress': 112,
    'mpc': 113,
    'ipOverAtm': 114,
    'iso88025Fiber': 115,
    'tdlc': 116,
    'gigabitEthernet': 117,
    'hdlc': 118,
    'lapf': 119,
    'v37': 120,
    'x25mlp': 121,
    'x25huntGroup': 122,
    'transpHdlc': 123,
    'interleave': 124,
    'fast': 125,
    'ip': 126,
    'docsCableMaclayer': 127,
    'docsCableDownstream': 128,
    'docsCableUpstream': 129,
    'a12MppSwitch': 130,
    'tunnel': 131,
    'coffee': 132,
    'ces': 133,
    'atmSubInterface': 134,
    'l2vlan': 135,
    'l3ipvlan': 136,
    'l3ipxvlan': 137,
    'digitalPowerline': 138,
    'mediaMailOverIp': 139,
    'dtm': 140,
    'dcn': 141,
    'ipForward': 142,
    'msdsl': 143,
    'ieee1394': 144,
    'if-gsn': 145,
    'dvbRccMacLayer': 146,
    'dvbRccDownstream': 147,
    'dvbRccUpstream': 148,
    'atmVirtual': 149,
    'mplsTunnel': 150,
    'srp': 151,
    'voiceOverAtm': 152,
    'voiceOverFrameRelay': 153,
    'idsl': 154,
    'compositeLink': 155,
    'ss7SigLink': 156,
    'propWirelessP2P': 157,
    'frForward': 158,
    'rfc1483': 159,
    'usb': 160,
    'ieee8023adLag': 161,
    'bgppolicyaccounting': 162,
    'frf16MfrBundle': 163,
    'h323Gatekeeper': 164,
    'h323Proxy': 165,
    'mpls': 166,
    'mfSigLink': 167,
    'hdsl2': 168,
    'shdsl': 169,
    'ds1FDL': 170,
    'pos': 171,
    'dvbAsiIn': 172,
    'dvbAsiOut': 173,
    'plc': 174,
    'nfas': 175,
    'tr008': 176,
    'gr303RDT': 177,
    'gr303IDT': 178,
    'isup': 179,
    'propDocsWirelessMaclayer': 180,
    'propDocsWirelessDownstream': 181,
    'propDocsWirelessUpstream': 182,
    'hiperlan2': 183,
    'propBWAp2Mp': 184,
    'sonetOverheadChannel': 185,
    'digitalWrapperOverheadChannel': 186,
    'aal2': 187,
    'radioMAC': 188,
    'atmRadio': 189,
    'imt': 190,
    'mvl': 191,
    'reachDSL': 192,
    'frDlciEndPt': 193,
    'atmVciEndPt': 194,
    'opticalChannel': 195,
    'opticalTransport': 196,
    'propAtm': 197,
    'voiceOverCable': 198,
    'infiniband': 199,
    'teLink': 200,
    'q2931': 201,
    'virtualTg': 202,
    'sipTg': 203,
    'sipSig': 204,
    'docsCableUpstreamChannel': 205,
    'econet': 206,
    'pon155': 207,
    'pon622': 208,
    'bridge': 209,
    'linegroup': 210,
    'voiceEMFGD': 211,
    'voiceFGDEANA': 212,
    'voiceDID': 213,
    'mpegTransport': 214,
    'sixToFour': 215,
    'gtp': 216,
    'pdnEtherLoop1': 217,
    'pdnEtherLoop2': 218,
    'opticalChannelGroup': 219,
    'homepna': 220,
    'gfp': 221,
    'ciscoISLvlan': 222,
    'actelisMetaLOOP': 223,
    'fcipLink': 224,
    'rpr': 225,
    'qam': 226,
    'lmp': 227,
    'cblVectaStar': 228,
    'docsCableMCmtsDownstream': 229,
    'adsl2': 230,
    'macSecControlledIF': 231,
    'macSecUncontrolledIF': 232,
    'aviciOpticalEther': 233,
    'atmbond': 234,
    'voiceFGDOS': 235,
    'mocaVersion1': 236,
    'ieee80216WMAN': 237,
    'adsl2plus': 238,
    'dvbRcsMacLayer': 239,
    'dvbTdm': 240,
    'dvbRcsTdma': 241,
    'x86Laps': 242,
    'wwanPP': 243,
    'wwanPP2': 244,
    'voiceEBS': 245,
    'ifPwType': 246,
    'ilan': 247,
    'pip': 248,
    'aluELP': 249,
    'gpon': 250,
    'vdsl2': 251,
    'capwapDot11Profile': 252,
    'capwapDot11Bss': 253,
    'capwapWtpVirtualRadio': 254,
    'bits': 255,
    'docsCableUpstreamRfPort': 256,
    'cableDownstreamRfPort': 257,
    'vmwareVirtualNic': 258,
    'ieee802154': 259,
})

"""
struct pcap_hdr_s {
        guint32 magic_number;   /* magic number */
        guint16 version_major;  /* major version number */
        guint16 version_minor;  /* minor version number */
        gint32  thiszone;       /* GMT to local correction */
        guint32 sigfigs;        /* accuracy of timestamps */
        guint32 snaplen;        /* max length of captured packets, in octets */
        guint32 network;        /* data link type */
} pcap_hdr_t;
typedef struct pcaprec_hdr_s {
        guint32 ts_sec;         /* timestamp seconds */
        guint32 ts_usec;        /* timestamp microseconds */
        guint32 incl_len;       /* number of octets of packet saved in file */
        guint32 orig_len;       /* actual length of packet */
} pcaprec_hdr_t;
"""

LINKTYPE_ETHERNET = 1
LINKTYPE_RAW = 101

PCAP_HEADER = struct.pack('<LHHLLLL', 0xa1b2c3d4, 2, 4, 0, 0, 65535, LINKTYPE_ETHERNET)
PACKET_HEADER = '<LLLL'
ETHERNET_HEADER = b'\xBE\xEF' * 6 + b'\x08\x00'

device_packet_struct = Struct(
    'header_length' / Int32ub,
    'header_version' / Byte,
    'packet_length' / Int32ub,
    'interface_type' / Byte,
    Padding(2),
    Padding(1),
    'protocol_family' / Int32ub,
    'frame_pre_length' / Int32ub,
    'frame_post_length' / Int32ub,
    'interface_name' / Padded(16, CString('utf8')),
    'pid' / Int32ul,
    'comm' / Padded(17, CString('utf8')),
    'svc' / Int32ub,
    'epid' / Int32ul,
    'ecomm' / Padded(17, CString('utf8')),
    'seconds' / Int32ub,
    'microseconds' / Int32ub,
    'data' / Bytes(this.packet_length)
)


class PcapdService(BaseService):
    """
    Starting iOS 5, apple added a remote virtual interface (RVI) facility that allows mirroring networks trafic from
    an iOS device. On macOS, the virtual interface can be enabled with the rvictl command. This script allows to use
    this service on other systems.
    """
    SERVICE_NAME = 'com.apple.pcapd'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    def watch(self, packets_count: int = -1, process: str = None):
        packet_index = 0
        while packet_index != packets_count:
            d = self.service.recv_plist()
            if not d:
                break

            packet = device_packet_struct.parse(d)

            if process is not None:
                if process != str(packet.pid) and process != packet.comm:
                    continue

            if not packet.frame_pre_length:
                # Add fake ethernet header for pdp packets.
                packet.data = ETHERNET_HEADER + packet.data
            packet.interface_type = INTERFACE_NAMES(packet.interface_type)
            packet.protocol_family = socket.AddressFamily(packet.protocol_family)

            yield packet

            packet_index += 1

    @staticmethod
    def write_to_pcap(out, packet_generator):
        out.write(PCAP_HEADER)
        for packet in packet_generator:
            length = len(packet.data)
            pkthdr = struct.pack(PACKET_HEADER, packet.seconds, packet.microseconds, length, length)
            data = pkthdr + packet.data
            out.write(data)



import binascii
import logging
import math
import struct
import time
from enum import Enum
from pymobiledevice3.exceptions import PyMobileDevice3Exception, IRecvNoDeviceConnectedError
from pymobiledevice3.irecv_devices import IRECV_DEVICES, IRecvDevice
from tqdm import trange
from typing import Optional
from usb.core import find, Device, USBError
from usb.util import get_string

USB_TIMEOUT = 10000
IRECV_TRANSFER_SIZE_RECOVERY = 0x8000
IRECV_TRANSFER_SIZE_DFU = 0x800


class Mode(Enum):
    RECOVERY_MODE_1 = 0x1280
    RECOVERY_MODE_2 = 0x1281
    RECOVERY_MODE_3 = 0x1282
    RECOVERY_MODE_4 = 0x1283
    WTF_MODE = 0x1222
    DFU_MODE = 0x1227

    @classmethod
    def has_value(cls, value):
        for m in cls:
            if value == m.value:
                return True
        return False

    @classmethod
    def get_mode_from_value(cls, value):
        """
        :rtype: Mode
        """
        for m in cls:
            if value == m.value:
                return m
        return None

    @property
    def is_recovery(self):
        return self not in (self.WTF_MODE, self.DFU_MODE)


CPFM_FLAG_SECURITY_MODE = 1 << 0
CPFM_FLAG_PRODUCTION_MODE = 1 << 1

IBOOT_FLAG_IMAGE4_AWARE = 1 << 2
IBOOT_FLAG_EFFECTIVE_SECURITY_MODE = 1 << 3
IBOOT_FLAG_EFFECTIVE_PRODUCTION_MODE = 1 << 4

APPLE_VENDOR_ID = 0x05AC

logger = logging.getLogger(__name__)


class IRecv:
    def __init__(self, ecid=None, timeout=0xffffffff, is_recovery=None):
        self.mode = None  # type: Optional[Mode]
        self._device_info = {}
        self._device = None  # type: Optional[Device]
        self._reinit(ecid=ecid, timeout=timeout, is_recovery=is_recovery)

    @property
    def ecid(self):
        return int(self._device_info['ECID'], 16)

    @property
    def ibfl(self):
        return int(self._device_info['IBFL'], 16)

    @property
    def chip_id(self):
        return int(self._device_info['CPID'], 16)

    @property
    def board_id(self):
        return int(self._device_info['BDID'], 16)

    @property
    def serial_number(self) -> int:
        return int(self._device_info['SRNM'], 16)

    @property
    def iboot_version(self) -> str:
        return self._device_info['SRTG']

    @property
    def is_image4_supported(self):
        return self.ibfl & IBOOT_FLAG_IMAGE4_AWARE

    @property
    def _irecv_device(self) -> IRecvDevice:
        for device in IRECV_DEVICES:
            if device.board_id == self.board_id and device.chip_id == self.chip_id:
                return device
        raise KeyError(f'failed to find device of: board_id: {self.board_id} chip_id: {self.chip_id}')

    @property
    def product_type(self):
        return self._irecv_device.product_type

    @property
    def hardware_model(self):
        return self._irecv_device.hardware_model

    @property
    def display_name(self):
        return self._irecv_device.display_name

    @property
    def status(self):
        return self.ctrl_transfer(0xa1, 3, data_or_wLength=b'\x00' * 6)[4]

    def set_interface_altsetting(self, interface=None, alternate_setting=None):
        logger.debug(f'set_interface_altsetting: {interface} {alternate_setting}')
        if interface == 1:
            self._device.set_interface_altsetting(interface=interface, alternate_setting=alternate_setting)

    def set_configuration(self, configuration=None):
        logger.debug(f'set_configuration: {configuration}')
        if self._device.get_active_configuration().bConfigurationValue != configuration:
            self._device.set_configuration(configuration=configuration)

    def ctrl_transfer(self, bmRequestType, bRequest, timeout=USB_TIMEOUT, **kwargs):
        return self._device.ctrl_transfer(bmRequestType, bRequest, timeout=timeout, **kwargs)

    def send_buffer(self, buf: bytes):
        packet_size = IRECV_TRANSFER_SIZE_RECOVERY if self.mode.is_recovery else IRECV_TRANSFER_SIZE_DFU

        # initiate transfer
        if self.mode.is_recovery:
            self.ctrl_transfer(0x41, 0)
        else:
            response = self.ctrl_transfer(0xa1, 5, data_or_wLength=1)
            state = response[0]
            logger.debug(f'irecv state: {state}')
            if state == 2:
                # DFU IDLE
                pass
            elif state == 10:
                self.ctrl_transfer(0x21, 4)
                raise PyMobileDevice3Exception('DFU ERROR, issuing CLRSTATUS')
            else:
                self.ctrl_transfer(0x21, 6)
                raise PyMobileDevice3Exception(f'Unexpected state {state}, issuing ABORT')

        crc = -1

        num_packets = math.ceil(len(buf) / packet_size)

        for offset in trange(0, len(buf), packet_size, dynamic_ncols=True):
            # Use bulk transfer for recovery mode and control transfer for DFU and WTF mode
            chunk = buf[offset:offset + packet_size]
            packet_index = offset // packet_size

            if self.mode.is_recovery:
                n = self._device.write(0x04, chunk, timeout=USB_TIMEOUT)
                if n != len(chunk):
                    raise IOError('failed to upload data')
            else:
                if offset + packet_size >= len(buf):
                    # last packet

                    # calculate crc of all sent data
                    crc = binascii.crc32(buf, crc)

                    # add crc of dfu_xbuf (salted value)
                    dfu_xbuf = bytearray([0xff, 0xff, 0xff, 0xff, 0xac, 0x05, 0x00, 0x01, 0x55, 0x46, 0x44, 0x10])
                    crc = binascii.crc32(dfu_xbuf, crc)

                    crc_chunk = dfu_xbuf + struct.pack('<I', crc)

                    if len(chunk) + 16 > packet_size:
                        # crc exceeds the max allowed packet size
                        self.ctrl_transfer(0x21, 1, wValue=packet_index, wIndex=0, data_or_wLength=chunk)
                        self.ctrl_transfer(0x21, 1, wValue=packet_index, wIndex=0, data_or_wLength=crc_chunk)
                    else:
                        self.ctrl_transfer(0x21, 1, wValue=packet_index, wIndex=0, data_or_wLength=chunk + crc_chunk)
                else:
                    self.ctrl_transfer(0x21, 1, wValue=packet_index, wIndex=0, data_or_wLength=chunk)

        if not self.mode.is_recovery:
            logger.debug('waiting for status == 5')
            while self.status != 5:
                time.sleep(1)

            self.ctrl_transfer(0x21, 1, wValue=num_packets, wIndex=0)

            for offset in range(2):
                # i know it's not used but idevicerestore does that also
                _ = self.status

            self.reset()

    def reset(self):
        try:
            logger.debug('resetting usb device')
            self._device.reset()
        except USBError:
            pass

        self._reinit(ecid=self.ecid)

    def send_command(self, cmd: str, timeout=USB_TIMEOUT, b_request=0):
        self._device.ctrl_transfer(0x40, b_request, 0, 0, cmd.encode() + b'\0', timeout=timeout)

    def getenv(self, name):
        try:
            self.send_command(f'getenv {name}')
        except USBError:
            return None
        return bytearray(self._device.ctrl_transfer(0xc0, 0, 0, 0, 255))

    def set_autoboot(self, enable: bool):
        self.send_command(f'setenv auto-boot {str(enable).lower()}')
        self.send_command('saveenv')

    def reboot(self):
        try:
            self.send_command('reboot')
        except USBError:
            pass

    def _reinit(self, ecid=None, timeout=0xffffffff, is_recovery=None):
        self._device = None
        self._device_info = {}
        self.mode = None
        self._find(ecid=ecid, timeout=timeout, is_recovery=is_recovery)
        if self._device is None:
            raise IRecvNoDeviceConnectedError('Failed to find a connected iDevice')
        self._populate_device_info()

        self.ap_nonce = self._copy_nonce_with_tag('NONC')
        self.sep_nonce = self._copy_nonce_with_tag('SNON')

        self.set_configuration(1)

        if self.mode.is_recovery:
            self.set_interface_altsetting(0, 0)
            if self.mode.value > Mode.RECOVERY_MODE_2.value:
                self.set_interface_altsetting(1, 1)
        else:
            self.set_interface_altsetting(0, 0)

    def _copy_nonce_with_tag(self, tag):
        return binascii.unhexlify(get_string(self._device, 1).split(f'{tag}:')[1].split(' ')[0])

    def _find(self, ecid=None, timeout=0xffffffff, is_recovery=None):
        start = time.time()
        end = start + timeout
        while (self._device is None) and (time.time() < end):
            for device in find(find_all=True):
                try:
                    if device.manufacturer is None:
                        continue
                    if not device.manufacturer.startswith('Apple'):
                        continue

                    mode = Mode.get_mode_from_value(device.idProduct)
                    if mode is None:
                        # not one of Apple's special modes
                        continue

                    if is_recovery is not None and mode.is_recovery != is_recovery:
                        continue

                    if self._device is not None:
                        raise Exception('More then one connected device was found connected in recovery mode')
                    self._device = device
                    self.mode = mode
                    self._populate_device_info()

                    if ecid is not None:
                        found_ecid = int(self._device_info['ECID'], 16)
                        if found_ecid != ecid:
                            # wrong device - move on
                            self._device = None
                            continue
                except ValueError:
                    continue

    def _populate_device_info(self):
        for component in self._device.serial_number.split(' '):
            k, v = component.split(':')
            if k in ('SRNM', 'SRTG') and '[' in v:
                # trim the `[]`
                v = v[1:-1]
            self._device_info[k] = v

    def __str__(self):
        return str(self._device_info)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self._device


def main():
    print(IRecv())


if __name__ == '__main__':
    main()



import logging
import os
import plistlib
import traceback

import IPython
import click
from pygments import highlight, lexers, formatters

from pymobiledevice3 import usbmux
from pymobiledevice3.cli.cli_common import print_json, set_verbosity
from pymobiledevice3.exceptions import IncorrectModeError
from pymobiledevice3.irecv import IRecv
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.restore.device import Device
from pymobiledevice3.restore.recovery import Recovery, Behavior
from pymobiledevice3.restore.restore import Restore
from pymobiledevice3.services.diagnostics import DiagnosticsService

SHELL_USAGE = """
# use `irecv` variable to access Restore mode API
# for example:
print(irecv.getenv('build-version'))
"""

logger = logging.getLogger(__name__)


class Command(click.Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params[:0] = [
            click.Option(('device', '--ecid'), type=click.INT, callback=self.device),
            click.Option(('verbosity', '-v', '--verbose'), count=True, callback=set_verbosity, expose_value=False),
        ]

    @staticmethod
    def device(ctx, param, value):
        if '_PYMOBILEDEVICE3_COMPLETE' in os.environ:
            # prevent lockdown connection establishment when in autocomplete mode
            return

        ecid = value
        logger.debug('searching among connected devices via lockdownd')
        for device in usbmux.list_devices():
            try:
                lockdown = LockdownClient(serial=device.serial)
            except IncorrectModeError:
                continue
            if (ecid is None) or (lockdown.ecid == value):
                logger.debug('found device')
                return lockdown
            else:
                continue
        logger.debug('waiting for device to be available in Recovery mode')
        return IRecv(ecid=ecid)


@click.group()
def cli():
    """ cli """
    pass


@cli.group()
def restore():
    """ restore options """
    pass


@restore.command('shell', cls=Command)
def restore_shell(device):
    """ create an IPython shell for interacting with iBoot """
    IPython.embed(
        header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.TerminalTrueColorFormatter(style='native')),
        user_ns={
            'irecv': device,
        })


@restore.command('enter', cls=Command)
def restore_enter(device):
    """ enter Recovery mode """
    if isinstance(device, LockdownClient):
        device.enter_recovery()


@restore.command('exit')
def restore_exit():
    """ exit Recovery mode """
    irecv = IRecv()
    irecv.set_autoboot(True)
    irecv.reboot()


@restore.command('restart', cls=Command)
def restore_restart(device):
    """ restarts device """
    if isinstance(device, LockdownClient):
        with DiagnosticsService(device) as diagnostics:
            diagnostics.restart()
    else:
        device.reboot()


@restore.command('tss', cls=Command)
@click.argument('ipsw', type=click.File('rb'))
@click.argument('out', type=click.File('wb'), required=False)
@click.option('--color/--no-color', default=True)
def restore_tss(device, ipsw, out, color):
    """ query SHSH blobs """
    lockdown = None
    irecv = None
    if isinstance(device, LockdownClient):
        lockdown = device
    elif isinstance(device, IRecv):
        irecv = device

    device = Device(lockdown=lockdown, irecv=irecv)
    tss = Recovery(ipsw, device).fetch_tss_record()
    if out:
        plistlib.dump(tss, out)
    print_json(tss, colored=color)


@restore.command('ramdisk', cls=Command)
@click.argument('ipsw', type=click.File('rb'))
@click.option('--tss', type=click.File('rb'))
def restore_ramdisk(device, ipsw, tss):
    """ don't perform an actual restore. just enter the update ramdisk """
    if tss:
        tss = plistlib.load(tss)

    lockdown = None
    irecv = None
    if isinstance(device, LockdownClient):
        lockdown = device
    elif isinstance(device, IRecv):
        irecv = device
    device = Device(lockdown=lockdown, irecv=irecv)
    Recovery(ipsw, device, tss=tss).boot_ramdisk()


@restore.command('update', cls=Command)
@click.argument('ipsw', type=click.File('rb'))
@click.option('--tss', type=click.File('rb'))
@click.option('--erase', is_flag=True, help='use the Erase BuildIdentity (full factory-reset)')
@click.option('--ignore-fdr', is_flag=True, help='only establish an FDR service connection, but don\'t proxy any '
                                                 'traffic')
def restore_update(device, ipsw, tss, erase, ignore_fdr):
    """ perform an upgrade """
    if tss:
        tss = plistlib.load(tss)

    lockdown = None
    irecv = None
    if isinstance(device, LockdownClient):
        lockdown = device
    elif isinstance(device, IRecv):
        irecv = device
    device = Device(lockdown=lockdown, irecv=irecv)

    behavior = Behavior.Update
    if erase:
        behavior = Behavior.Erase

    try:
        Restore(ipsw, device, tss=tss, behavior=behavior, ignore_fdr=ignore_fdr).update()
    except Exception:
        # click may "swallow" several exception types so we try to catch them all here
        traceback.print_exc()
        raise



import logging

import click

from pymobiledevice3.cli.cli_common import Command
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.resources.firmware_notifications import get_notifications
from pymobiledevice3.services.notification_proxy import NotificationProxyService

logger = logging.getLogger(__name__)


@click.group()
def cli():
    """ notification options """
    pass


@cli.group()
def notification():
    """ notification options """
    pass


@notification.command(cls=Command)
@click.argument('names', nargs=-1)
@click.option('--insecure', is_flag=True, help='use the insecure relay meant for untrusted clients instead')
def post(lockdown: LockdownClient, names, insecure):
    """ API for notify_post(). """
    service = NotificationProxyService(lockdown=lockdown, insecure=insecure)
    for name in names:
        service.notify_post(name)


@notification.command(cls=Command)
@click.argument('names', nargs=-1)
@click.option('--insecure', is_flag=True, help='use the insecure relay meant for untrusted clients instead')
def observe(lockdown: LockdownClient, names, insecure):
    """ API for notify_register_dispatch(). """
    service = NotificationProxyService(lockdown=lockdown, insecure=insecure)
    for name in names:
        service.notify_register_dispatch(name)

    for event in service.receive_notification():
        logger.info(event)


@notification.command('observe-all', cls=Command)
@click.option('--insecure', is_flag=True, help='use the insecure relay meant for untrusted clients instead')
def observe_all(lockdown: LockdownClient, insecure):
    """ attempt to observe all builtin firmware notifications. """
    service = NotificationProxyService(lockdown=lockdown, insecure=insecure)
    for notification in get_notifications():
        service.notify_register_dispatch(notification)

    for event in service.receive_notification():
        logger.info(event)



from collections import UserDict
from typing import List, Mapping

from cached_property import cached_property
from pymobiledevice3.restore.ipsw.component import Component


class BuildIdentity(UserDict):
    def __init__(self, build_manifest, data):
        super().__init__(data)
        self.build_manifest = build_manifest

    @cached_property
    def device_class(self):
        return self['Info']['DeviceClass'].lower()

    @cached_property
    def restore_behavior(self):
        return self['Info'].get('RestoreBehavior')

    @cached_property
    def variant(self):
        return self['Info'].get('Variant')

    @cached_property
    def macos_variant(self):
        return self['Info'].get('MacOSVariant')

    @cached_property
    def manifest(self):
        return self['Manifest']

    @cached_property
    def minimum_system_partition(self):
        return self['Info'].get('MinimumSystemPartition')

    def get_component_path(self, component: str):
        return self.manifest[component]['Info']['Path']

    def has_component(self, name: str):
        return name in self.manifest

    def get_component(self, name: str, **args) -> Component:
        return Component(self, name, **args)

    def populate_tss_request_parameters(self, parameters: Mapping, additional_keys: List[str] = None):
        """ equivalent to idevicerestore:tss_parameters_add_from_manifest """
        key_list = ['ApBoardID', 'ApChipID']
        if additional_keys is None:
            key_list += ['UniqueBuildID', 'Ap,OSLongVersion', 'ApChipID', 'ApBoardID', 'ApSecurityDomain',
                         'BMU,BoardID', 'BMU,ChipID', 'BbChipID', 'BbProvisioningManifestKeyHash',
                         'BbActivationManifestKeyHash', 'BbCalibrationManifestKeyHash',
                         'BbFactoryActivationManifestKeyHash', 'BbFDRSecurityKeyHash', 'BbSkeyId', 'SE,ChipID',
                         'Savage,ChipID', 'Savage,PatchEpoch', 'Yonkers,BoardID', 'Yonkers,ChipID',
                         'Yonkers,PatchEpoch', 'Rap,BoardID', 'Rap,ChipID', 'Rap,SecurityDomain', 'Baobab,BoardID',
                         'Baobab,ChipID', 'Baobab,ManifestEpoch', 'Baobab,SecurityDomain', 'eUICC,ChipID',
                         'PearlCertificationRootPub', 'NeRDEpoch', 'Timer,BoardID,1', 'Timer,BoardID,2',
                         'Timer,ChipID,1', 'Timer,ChipID,2', 'Timer,SecurityDomain,1', 'Timer,SecurityDomain,2',
                         'Manifest', ]
        else:
            key_list += additional_keys

        for k in key_list:
            try:
                v = self[k]
                if isinstance(v, str) and v.startswith('0x'):
                    v = int(v, 16)
                parameters[k] = v
            except KeyError:
                pass

        requires_uid_mode = self['Info'].get('RequiresUIDMode')
        if requires_uid_mode is not None:
            parameters['RequiresUIDMode'] = requires_uid_mode



#!/usr/bin/env python3
import typing

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.base_service import BaseService


class CompanionProxyService(BaseService):
    SERVICE_NAME = 'com.apple.companion_proxy'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    def list(self):
        service = self.lockdown.start_service(self.SERVICE_NAME)
        return service.send_recv_plist({'Command': 'GetDeviceRegistry'}).get('PairedDevicesArray', [])

    def listen_for_devices(self):
        service = self.lockdown.start_service(self.SERVICE_NAME)
        service.send_plist({'Command': 'StartListeningForDevices'})
        while True:
            yield service.recv_plist()

    def get_value(self, udid: str, key: str):
        service = self.lockdown.start_service(self.SERVICE_NAME)
        response = service.send_recv_plist({'Command': 'GetValueFromRegistry',
                                            'GetValueGizmoUDIDKey': udid,
                                            'GetValueKeyKey': key})

        value = response.get('RetrievedValueDictionary')
        if value is not None:
            return value

        error = response.get('Error')
        raise PyMobileDevice3Exception(error)

    def start_forwarding_service_port(self, remote_port: int, service_name: str = None, options: typing.Mapping = None):
        service = self.lockdown.start_service(self.SERVICE_NAME)

        request = {'Command': 'StartForwardingServicePort',
                   'GizmoRemotePortNumber': remote_port,
                   'IsServiceLowPriority': False,
                   'PreferWifi': False}

        if service_name is not None:
            request['ForwardedServiceName'] = service_name

        if options is not None:
            request.update(options)

        return service.send_recv_plist(request).get('CompanionProxyServicePort')

    def stop_forwarding_service_port(self, remote_port: int):
        service = self.lockdown.start_service(self.SERVICE_NAME)

        request = {'Command': 'StopForwardingServicePort',
                   'GizmoRemotePortNumber': remote_port}

        return service.send_recv_plist(request)



__all__ = [
    'PyMobileDevice3Exception', 'DeviceVersionNotSupportedError', 'IncorrectModeError', 'DeviceVersionFormatError',
    'ConnectionFailedError', 'NotTrustedError', 'PairingError', 'NotPairedError', 'CannotStopSessionError',
    'PasswordRequiredError', 'StartServiceError', 'FatalPairingError', 'NoDeviceConnectedError', 'MuxException',
    'MuxVersionError', 'ArgumentError', 'AfcException', 'AfcFileNotFoundError', 'DvtException', 'DvtDirListError',
    'NotMountedError', 'AlreadyMountedError', 'UnsupportedCommandError', 'ExtractingStackshotError',
    'ConnectionTerminatedError', 'WirError', 'WebInspectorNotEnabled', 'RemoteAutomationNotEnabled',
    'ArbitrationError', 'NoSuchBuildIdentityError', 'InternalError', 'DeveloperModeIsNotEnabledError',
    'DeviceAlreadyInUseError', 'LockdownError', 'PairingDialogResponsePendingError', 'UserDeniedPairingError',
    'InvalidHostIDError', 'SetProhibitedError', 'MissingValueError', 'PasscodeRequiredError', 'AmfiError',
    'DeviceHasPasscodeSetError', 'NotificationTimeoutError', 'DeveloperModeError', 'ProfileError',
    'UsbmuxConnectionError', 'IRecvError', 'IRecvNoDeviceConnectedError', 'NoDeviceSelectedError',
    'MessageNotSupportedError',
]


class PyMobileDevice3Exception(Exception):
    pass


class DeviceVersionNotSupportedError(PyMobileDevice3Exception):
    pass


class IncorrectModeError(PyMobileDevice3Exception):
    pass


class DeviceVersionFormatError(PyMobileDevice3Exception):
    pass


class ConnectionFailedError(PyMobileDevice3Exception):
    pass


class NotTrustedError(PyMobileDevice3Exception):
    pass


class PairingError(PyMobileDevice3Exception):
    pass


class NotPairedError(PyMobileDevice3Exception):
    pass


class CannotStopSessionError(PyMobileDevice3Exception):
    pass


class PasswordRequiredError(PairingError):
    pass


class StartServiceError(PyMobileDevice3Exception):
    pass


class FatalPairingError(PyMobileDevice3Exception):
    pass


class NoDeviceConnectedError(PyMobileDevice3Exception):
    pass


class MuxException(PyMobileDevice3Exception):
    pass


class MuxVersionError(MuxException):
    pass


class ArgumentError(PyMobileDevice3Exception):
    pass


class AfcException(PyMobileDevice3Exception, OSError):
    def __init__(self, message, status):
        OSError.__init__(self, status, message)
        self.status = status


class AfcFileNotFoundError(AfcException):
    pass


class DvtException(PyMobileDevice3Exception):
    """ Domain exception for DVT operations. """
    pass


class UnrecognizedSelectorError(DvtException):
    """ Attempted to call an unrecognized selector from DVT. """
    pass


class DvtDirListError(DvtException):
    """ Raise when directory listing fails. """
    pass


class NotMountedError(PyMobileDevice3Exception):
    """ Given image for umount wasn't mounted in the first place """
    pass


class AlreadyMountedError(PyMobileDevice3Exception):
    """ Given image for mount has already been mounted in the first place """
    pass


class UnsupportedCommandError(PyMobileDevice3Exception):
    """ Given command isn't supported for this iOS version """
    pass


class ExtractingStackshotError(PyMobileDevice3Exception):
    """ Raise when stackshot is not received in the core profile session. """
    pass


class ConnectionTerminatedError(PyMobileDevice3Exception):
    """ Raise when a connection is terminated abruptly. """
    pass


class WebInspectorNotEnabled(PyMobileDevice3Exception):
    """ Raise when Web Inspector is not enabled. """
    pass


class RemoteAutomationNotEnabled(PyMobileDevice3Exception):
    """ Raise when Web Inspector remote automation is not enabled. """
    pass


class WirError(PyMobileDevice3Exception):
    """ Raise when Webinspector WIR command fails. """
    pass


class InternalError(PyMobileDevice3Exception):
    """ Some internal Apple error """
    pass


class ArbitrationError(PyMobileDevice3Exception):
    """ Arbitration failed """
    pass


class DeviceAlreadyInUseError(ArbitrationError):
    """ Device is already checked-in by someone """

    @property
    def message(self):
        return self.args[0].get('message')

    @property
    def owner(self):
        return self.args[0].get('owner')

    @property
    def result(self):
        return self.args[0].get('result')


class NoSuchBuildIdentityError(PyMobileDevice3Exception):
    """ Failed to find BuildIdentity with desired parameters """
    pass


class DeveloperModeIsNotEnabledError(PyMobileDevice3Exception):
    """ Raise when mounting failed because developer mode is not enabled. """
    pass


class DeveloperModeError(PyMobileDevice3Exception):
    """ Raise when amfid failed to enable developer mode. """
    pass


class LockdownError(PyMobileDevice3Exception):
    """ lockdown general error """
    pass


class SetProhibitedError(LockdownError):
    pass


class PairingDialogResponsePendingError(PairingError):
    """ User hasn't yet confirmed the device is trusted """
    pass


class UserDeniedPairingError(PairingError):
    pass


class InvalidHostIDError(PairingError):
    pass


class MissingValueError(LockdownError):
    """ raised when attempting to query non-existent domain/key """
    pass


class PasscodeRequiredError(LockdownError):
    """ passcode must be present for this action """
    pass


class AmfiError(PyMobileDevice3Exception):
    pass


class DeviceHasPasscodeSetError(AmfiError):
    pass


class NotificationTimeoutError(PyMobileDevice3Exception, TimeoutError):
    pass


class ProfileError(PyMobileDevice3Exception):
    pass


class UsbmuxConnectionError(PyMobileDevice3Exception):
    """ error connecting to usbmuxd socket """
    pass


class IRecvError(PyMobileDevice3Exception):
    pass


class IRecvNoDeviceConnectedError(IRecvError):
    pass


class NoDeviceSelectedError(PyMobileDevice3Exception):
    pass


class MessageNotSupportedError(PyMobileDevice3Exception):
    pass



import click

from pymobiledevice3.cli.cli_common import Command
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.afc import AfcService, AfcShell


@click.group()
def cli():
    """ apps cli """
    pass


@cli.group()
def afc():
    """ FileSystem utils """
    pass


@afc.command('shell', cls=Command)
def afc_shell(lockdown: LockdownClient):
    """ open an AFC shell rooted at /var/mobile/Media """
    AfcShell(lockdown=lockdown, service_name='com.apple.afc').cmdloop()


@afc.command('pull', cls=Command)
@click.argument('remote_file', type=click.Path(exists=False))
@click.argument('local_file', type=click.File('wb'))
def afc_pull(lockdown: LockdownClient, remote_file, local_file):
    """ pull remote file from /var/mobile/Media """
    local_file.write(AfcService(lockdown=lockdown).get_file_contents(remote_file))


@afc.command('push', cls=Command)
@click.argument('local_file', type=click.File('rb'))
@click.argument('remote_file', type=click.Path(exists=False))
def afc_push(lockdown: LockdownClient, local_file, remote_file):
    """ push local file into /var/mobile/Media """
    AfcService(lockdown=lockdown).set_file_contents(remote_file, local_file.read())


@afc.command('ls', cls=Command)
@click.argument('remote_file', type=click.Path(exists=False))
@click.option('-r', '--recursive', is_flag=True)
def afc_ls(lockdown: LockdownClient, remote_file, recursive):
    """ perform a dirlist rooted at /var/mobile/Media """
    for path in AfcService(lockdown=lockdown).dirlist(remote_file, -1 if recursive else 1):
        print(path)


@afc.command('rm', cls=Command)
@click.argument('remote_file', type=click.Path(exists=False))
def afc_rm(lockdown: LockdownClient, remote_file):
    """ remove a file rooted at /var/mobile/Media """
    AfcService(lockdown=lockdown).rm(remote_file)
